from typing import Dict

from fastf1.plotting._constants import (  # noqa: F401, unused import used through globals()
    season2018,
    season2019,
    season2020,
    season2021,
    season2022,
    season2023,
    season2024
)
from fastf1.plotting._constants.base import BaseSeason


Constants: Dict[str, BaseSeason] = dict()

for year in range(2018, 2025):
    season = globals()[f"season{year}"]
    Constants[str(year)] = BaseSeason(CompoundColors=season.CompoundColors,
                                      Teams=season.Teams)


# Deprecated, will be removed for 2025
LEGACY_TEAM_COLORS = {
    'mercedes': '#00d2be', 'ferrari': '#dc0000',
    'red bull': '#fcd700', 'mclaren': '#ff8700',
    'alpine': '#fe86bc', 'aston martin': '#006f62',
    'sauber': '#00e701', 'visa rb': '#1634cb',
    'haas': '#ffffff', 'williams': '#00a0dd'
}


LEGACY_TEAM_TRANSLATE: Dict[str, str] = {
    'MER': 'mercedes',
    'FER': 'ferrari',
    'RBR': 'red bull',
    'MCL': 'mclaren',
    'APN': 'alpine',
    'AMR': 'aston martin',
    'SAU': 'sauber',
    'RB': 'rb',
    'HAA': 'haas',
    'WIL': 'williams'
}


LEGACY_DRIVER_COLORS: Dict[str, str] = {
    "valtteri bottas": "#00e701",
    "zhou guanyu": "#008d01",
    "theo pourchaire": "#004601",

    "nyck de vries": "#1e3d61",
    "yuki tsunoda": "#356cac",
    "daniel ricciardo": "#2b4562",
    "liam lawson": "#2b4562",
    "isack hadjar": "#1e6176",
    "ayumu iwasa": "#1e6176",

    "pierre gasly": "#fe86bc",
    "esteban ocon": "#ff117c",
    "jack doohan": "#894667",

    "fernando alonso": "#006f62",
    "lance stroll": "#00413b",
    "felipe drugovich": "#2f9b90",

    "charles leclerc": "#dc0000",
    "carlos sainz": "#ff8181",
    "robert shwartzman": "#9c0000",
    "oliver bearman": "#c40000",

    "kevin magnussen": "#ffffff",
    "nico hulkenberg": "#cacaca",

    "oscar piastri": "#ff8700",
    "lando norris": "#eeb370",
    "pato oward": "#ee6d3a",

    "lewis hamilton": "#00d2be",
    "george russell": "#24ffff",
    "frederik vesti": "#00a6ff",

    "max verstappen": "#fcd700",
    "sergio perez": "#ffec7b",
    "jake dennis": "#907400",

    "alexander albon": "#005aff",
    "logan sargeant": "#012564",
    "zak osullivan": "#1b3d97",
}


LEGACY_DRIVER_TRANSLATE: Dict[str, str] = {
    'LEC': 'charles leclerc', 'SAI': 'carlos sainz',
    'SHW': 'robert shwartzman',
    'VER': 'max verstappen', 'PER': 'sergio perez',
    'DEN': 'jake dennis',
    'PIA': 'oscar piastri', 'NOR': 'lando norris',
    'OWA': 'pato oward',
    'GAS': 'pierre gasly', 'OCO': 'esteban ocon',
    'DOO': 'jack doohan',
    'BOT': 'valtteri bottas', 'ZHO': 'zhou guanyu',
    'POU': 'theo pourchaire',
    'DEV': 'nyck de vries', 'TSU': 'yuki tsunoda',
    'RIC': 'daniel ricciardo', 'LAW': 'liam lawson',
    'HAD': 'isack hadjar', 'IWA': 'ayumu iwasa',
    'MAG': 'kevin magnussen', 'HUL': 'nico hulkenberg',
    'BEA': 'oliver bearman',
    'ALO': 'fernando alonso', 'STR': 'lance stroll',
    'DRU': 'felipe drugovich',
    'HAM': 'lewis hamilton', 'RUS': 'george russell',
    'VES': 'frederik vesti',
    'ALB': 'alexander albon', 'SAR': 'logan sargeant',
    'OSU': 'zak osullivan'}



import unicodedata
from typing import (
    Dict,
    List,
    Sequence,
    TypeVar
)

from rapidfuzz import fuzz

from fastf1.logger import get_logger


_logger = get_logger(__package__)


class _Driver:
    value: str = ''
    normalized_value: str = ''
    abbreviation: str = ''
    team: "_Team"


class _Team:
    value: str = ''
    normalized_value: str = ''

    def __init__(self):
        super().__init__()
        self.drivers: List["_Driver"] = list()


class _DriverTeamMapping:
    def __init__(
            self,
            year: str,
            teams: List[_Team],
    ):
        self.year = year
        self.teams = teams

        self.drivers_by_normalized: Dict[str, _Driver] = dict()
        self.drivers_by_abbreviation: Dict[str, _Driver] = dict()
        self.teams_by_normalized: Dict[str, _Team] = dict()

        for team in teams:
            for driver in team.drivers:
                self.drivers_by_normalized[driver.normalized_value] = driver
                self.drivers_by_abbreviation[driver.abbreviation] = driver
            self.teams_by_normalized[team.normalized_value] = team


S = TypeVar('S', bound=str)


def _fuzzy_matcher(identifier: str, reference: Sequence[S]) -> S:
    # do fuzzy string matching
    key_ratios = list()
    for existing_key in reference:
        ratio = fuzz.ratio(identifier, existing_key)
        key_ratios.append((ratio, existing_key))
    key_ratios.sort(reverse=True)
    if ((key_ratios[0][0] < 35)
            or (key_ratios[0][0] / key_ratios[1][0] < 1.2)):
        # ensure that the best match has a minimum accuracy (35 out of
        # 100) and that it has a minimum confidence (at least 20%
        # better than second best)
        raise KeyError
    if key_ratios[0][0] != 100:
        _logger.warning(
            ("Correcting invalid user input "
             f"'{identifier}' to '{key_ratios[0][1]}'."
             )
        )
    best_matched_key = key_ratios[0][1]
    return best_matched_key


def _normalize_string(name: str) -> str:
    # removes accents from a string and returns the closest possible
    # ascii representation (https://stackoverflow.com/a/518232)
    stripped = ''.join(c for c in unicodedata.normalize('NFD', name)
                       if unicodedata.category(c) != 'Mn')
    return stripped



"""Gear shifts on track
=======================

Plot which gear is being used at which point of the track
"""
##############################################################################
# Import FastF1 and load the data

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colormaps
from matplotlib.collections import LineCollection

import fastf1


session = fastf1.get_session(2021, 'Austrian Grand Prix', 'Q')
session.load()

lap = session.laps.pick_fastest()
tel = lap.get_telemetry()
# sphinx_gallery_defer_figures

##############################################################################
# Prepare the data for plotting by converting it to the appropriate numpy
# data types

x = np.array(tel['X'].values)
y = np.array(tel['Y'].values)

points = np.array([x, y]).T.reshape(-1, 1, 2)
segments = np.concatenate([points[:-1], points[1:]], axis=1)
gear = tel['nGear'].to_numpy().astype(float)
# sphinx_gallery_defer_figures

##############################################################################
# Create a line collection. Set a segmented colormap and normalize the plot
# to full integer values of the colormap

cmap = colormaps['Paired']
lc_comp = LineCollection(segments, norm=plt.Normalize(1, cmap.N+1), cmap=cmap)
lc_comp.set_array(gear)
lc_comp.set_linewidth(4)
# sphinx_gallery_defer_figures

##############################################################################
# Create the plot

plt.gca().add_collection(lc_comp)
plt.axis('equal')
plt.tick_params(labelleft=False, left=False, labelbottom=False, bottom=False)

title = plt.suptitle(
    f"Fastest Lap Gear Shift Visualization\n"
    f"{lap['Driver']} - {session.event['EventName']} {session.event.year}"
)
# sphinx_gallery_defer_figures


##############################################################################
# Add a colorbar to the plot. Shift the colorbar ticks by +0.5 so that they
# are centered for each color segment.

cbar = plt.colorbar(mappable=lc_comp, label="Gear",
                    boundaries=np.arange(1, 10))
cbar.set_ticks(np.arange(1.5, 9.5))
cbar.set_ticklabels(np.arange(1, 9))


plt.show()



import datetime
import re
from typing import Optional

from fastf1.logger import get_logger


logger = get_logger('ergast')


# ##############################################
# ### functions for date and time conversion ###

_time_string_matcher = re.compile(
    r'(\d{1,2}:)?(\d{1,2}:)?(\d{1,2})(\.\d{1,6})?(Z|[+-]\d{2}:\d{2})?'
)
# matches [hh:][mm:]ss[.micros][Z | +-hh:mm] timestring


def date_from_ergast(d_str) -> Optional[datetime.datetime]:
    """Create a ``datetime.datetime`` object from a date stamp formatted
    like 'YYYY-MM-DD'."""
    try:
        return datetime.datetime.strptime(d_str, "%Y-%m-%d")
    except ValueError:
        logger.debug(f"Failed to parse date stamp '{d_str}' in Ergast"
                     f"response.")
        return None


def time_from_ergast(t_str) -> Optional[datetime.time]:
    """Create a ``datetime.time`` object from a string that is formatted
    mostly like a timestamp according to ISO 8601. The implementation here only
    implements a subset of ISO 8601 to work around some missing functionality
    in ``datetime.time.fromisoformat`` in older versions of Python.

    Support timestamp format:
    [hh:][mm:]ss[.micros][Z/+-hh:mm]
    """
    # Examples for not supported strings on Python 3.8 in
    # ``datetime.time.fromisoformat``
    #   - 12:34.2 -> not accepted because only on microsecond decimal is given
    #   - 12:34Z  -> 'Z' is not accepted as an alias of '+00:00' (utc)

    res = _time_string_matcher.match(t_str)

    if res is None:
        logger.debug(f"Failed to parse timestamp '{t_str}' in Ergast"
                     f"response.")
        return None
    elif res[1] and res[2] and res[3]:
        hour, minute, second = int(res[1][:-1]), int(res[2][:-1]), int(res[3])
    elif res[1] and res[3]:
        hour, minute, second = 0, int(res[1][:-1]), int(res[3])
    else:
        hour, minute, second = 0, 0, int(res[3])

    if res[4]:
        digits = res[4][1:]
        microsecond = int(digits) * 10 ** (6 - len(digits))
    else:
        microsecond = 0

    if res[5] == 'Z':
        tzinfo = datetime.timezone.utc
    elif res[5]:
        tzinfo = datetime.datetime.strptime(res[5], "%z").tzinfo
    else:
        tzinfo = None

    try:
        return datetime.time(hour=hour, minute=minute, second=second,
                             microsecond=microsecond, tzinfo=tzinfo)
    except ValueError:
        logger.debug(f"Failed to parse timestamp '{t_str}' in Ergast"
                     f"response.")
        return None


def timedelta_from_ergast(t_str) -> Optional[datetime.timedelta]:
    """Create a ``datetime.timedelta`` object from a string that is formatted
    [+/-][hh:][mm:]ss[.micros], where all parts except for seconds are
    optional.
    """
    if t_str.startswith('-'):
        sign = -1
        t_str = t_str.strip('-')
    else:
        sign = 1
        t_str = t_str.strip('+')

    pseudo_time = time_from_ergast(t_str)
    if pseudo_time is not None:
        return sign * datetime.timedelta(hours=pseudo_time.hour,
                                         minutes=pseudo_time.minute,
                                         seconds=pseudo_time.second,
                                         microseconds=pseudo_time.microsecond)
    return None


def save_int(i_str) -> int:
    """
    Create an ``int`` object from a string that is formatted like an
    integer. In cases where the input string is not a valid integer,
    return -1. See #432
    """
    # Match pure integer strings, e.g.
    #   - '1234' -> 1234
    if re.match(r'^[+-]?\d+$', i_str):
        return int(i_str)

    # Otherwise, return -1. A notable example is #432: 1954 British GP, where
    # the racing time (in milliseconds) for Mike Hawthorn is an empty string
    #   - '' -> -1
    else:
        return -1


def save_float(f_str) -> float:
    """Create a ``float`` object from a string that is formatted like a float.
    In cases where the input string isn't a valid float, return nan
    """
    # Match pure float strings, e.g.
    #   - '1234.5678' -> 1234.5678
    if re.match(r'^[+-]?([0-9]+([.][0-9]*)?|[.][0-9]+)$', f_str):
        return float(f_str)

    # Otherwise, return np.nan
    else:
        return float('nan')


# ########################################################
# ### functions for flattening of ergast response data ###


def _flatten_by_rename(nested: dict, category: dict, flat: dict, *,
                       cast: bool = True, rename: bool = True):
    """:meta private:
    Iterate over all values on the current level, rename them and
    add them to the flattened result dict. This is the default operation that
    is used for most Ergast responses.

    Values that are not defined by category will be skipped and are not added
    to the flattened result.

    This function operates inplace on 'nested' and 'flat'.
    """
    for name, mapping in category['map'].items():
        if name not in nested:
            continue

        value = nested[name]
        if cast:
            value = mapping['type'](value)

        if rename:
            flat[mapping['name']] = value
        else:
            flat[name] = value


def _flatten_inline_list_of_dicts(nested: list, category: dict, flat: dict, *,
                                  cast: bool = True, rename: bool = True):
    """:meta private:
    The current level is a single list of dictionaries, iterate over them and
    convert from a list of dictionaries::

        [
            {"constructorId": "mclaren", ... },
            {"constructorId": "mercedes", ... },
            ...
        ]

    to a dictionary of lists::

        {"constructorIds": ["mclaren", "mercedes", ...], ...}

    This structure can then be included in the flattened result.

    For comparison, "normal" flattening returns a dictionary of strings,
    numbers, ... but in this case, the most reasonable way is to create arrays
    for all the individual values instead.

    Multi-mapping (one entry in the nested data mapped to multiple entries
    in the flat data) is supported.
    """
    # iterate over all values on the current level , join and rename them and
    # add the resulting lists to the flattened result dict
    for name, mapping in category['map'].items():
        # generate list of values for each mapping
        joined = list()
        for item in nested:
            if name not in item:
                continue

            value = item[name]
            if cast:
                value = mapping['type'](value)
            joined.append(value)
        if joined:
            if rename:
                flat[mapping['name']] = joined
            else:
                flat[name] = joined


def _lap_timings_flatten_by_rename(nested: dict, category: dict, flat: dict, *,
                                   cast: bool = True, rename: bool = True):
    """:meta private:
    Wrapper for :func:`flatten_by_rename` especially for lap timings.
    This function additionally directly integrates the subkey 'Timings' into
    the flattened results and converts the 'number' key to a list of value
    to match the values from 'Timings'.
    """
    # apply the normal flatten_by_rename function
    _flatten_by_rename(nested, category, flat, cast=cast, rename=rename)

    # pop the 'Timings' subcategory from the nested object to process it
    # here on this level already
    subcontent = nested.pop('Timings')
    # call its conversion method on its content to enable renaming and casting
    # by doing so, the subcontent is already directly integrated into flat
    Timings['method'](subcontent, Timings, flat, cast=cast, rename=rename)
    # the subcontent is a list of values for each key while 'number' is a
    # single value; therefore 'number' is augmented to a list of correct length
    flat['number'] = [flat['number'], ] * len(flat['driverId'])


# ################################
# ### root category finalizers ###

def _merge_dicts_of_lists(data):
    """:meta-private:
    Transform a list of equally keyed dictionaries that only contain lists into
    a single dictionary containing these list joined together.

        [
            {'value' : [1, 2, 3], ...},
            {'value' : [4, 5, 6], ...},
            ...
        ]

    Transform to ::

        {'value' : [1, 2, 3, 4, 5, 6, ...], ...},
    """
    if len(data) <= 1:
        return data[0]

    for _ in range(len(data) - 1):
        _tmp = data.pop(1)
        for key in data[0].keys():
            data[0][key].extend(_tmp.pop(key))

    return data[0]


# #####################################
# ### response subcategory elements ###

FirstPractice = {
    'name': 'FirstPractice',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'date': {'name': 'fp1Date', 'type': date_from_ergast},
            'time': {'name': 'fp1Time', 'type': time_from_ergast}},
    'sub': [],
    'finalize': None
}

SecondPractice = {
    'name': 'SecondPractice',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'date': {'name': 'fp2Date', 'type': date_from_ergast},
            'time': {'name': 'fp2Time', 'type': time_from_ergast}},
    'sub': [],
    'finalize': None
}

ThirdPractice = {
    'name': 'ThirdPractice',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'date': {'name': 'fp3Date', 'type': date_from_ergast},
            'time': {'name': 'fp3Time', 'type': time_from_ergast}},
    'sub': [],
    'finalize': None
}

Qualifying = {
    'name': 'Qualifying',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'date': {'name': 'qualifyingDate', 'type': date_from_ergast},
            'time': {'name': 'qualifyingTime', 'type': time_from_ergast}},
    'sub': [],
    'finalize': None
}

Sprint = {
    'name': 'Sprint',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'date': {'name': 'sprintDate', 'type': date_from_ergast},
            'time': {'name': 'sprintTime', 'type': time_from_ergast}},
    'sub': [],
    'finalize': None
}

TotalRaceTime = {
    'name': 'Time',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {
        'millis': {'name': 'totalRaceTimeMillis', 'type': save_int},
        'time': {'name': 'totalRaceTime', 'type': timedelta_from_ergast}
    },
    'sub': [],
    'finalize': None
}

FastestLapTime = {
    'name': 'Time',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {
        'millis': {'name': 'fastestLapTimeMillis', 'type': save_int},
        'time': {'name': 'fastestLapTime', 'type': timedelta_from_ergast}
    },
    'sub': [],
    'finalize': None
}

FastestLapAvgSpeed = {
    'name': 'AverageSpeed',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'units': {'name': 'fastestLapAvgSpeedUnits', 'type': str},
            'speed': {'name': 'fastestLapAvgSpeed', 'type': save_float}},
    'sub': [],
    'finalize': None
}

FastestLap = {
    'name': 'FastestLap',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'rank': {'name': 'fastestLapRank', 'type': save_int},
            'lap': {'name': 'fastestLapNumber', 'type': save_int}},
    'sub': [FastestLapTime, FastestLapAvgSpeed],
    'finalize': None
}

Driver = {
    'name': 'Driver',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'driverId': {'name': 'driverId', 'type': str},
            'permanentNumber': {'name': 'driverNumber', 'type': save_int},
            'code': {'name': 'driverCode', 'type': str},
            'url': {'name': 'driverUrl', 'type': str},
            'givenName': {'name': 'givenName', 'type': str},
            'familyName': {'name': 'familyName', 'type': str},
            'dateOfBirth': {'name': 'dateOfBirth', 'type': date_from_ergast},
            'nationality': {'name': 'driverNationality', 'type': str}},
    'sub': [],
    'finalize': None
}

Constructor = {
    'name': 'Constructor',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'constructorId': {'name': 'constructorId', 'type': str},
            'url': {'name': 'constructorUrl', 'type': str},
            'name': {'name': 'constructorName', 'type': str},
            'nationality': {'name': 'constructorNationality', 'type': str}},
    'sub': [],
    'finalize': None
}

ConstructorsInline = {
    # special case for where a list of constructors is given for a single
    # 'element' within the result (example: driver standings with potentially
    # multiple constructors per driver); these are reduced to a list of values
    # for each data field
    'name': 'Constructors',
    'type': list,
    'method': _flatten_inline_list_of_dicts,
    'map': {'constructorId': {'name': 'constructorIds', 'type': str},
            'url': {'name': 'constructorUrls', 'type': str},
            'name': {'name': 'constructorNames', 'type': str},
            'nationality': {'name': 'constructorNationalities', 'type': str}},
    'sub': [],
    'finalize': None
}

Location = {
    'name': 'Location',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'lat': {'name': 'lat', 'type': save_float},
            'long': {'name': 'long', 'type': save_float},
            'locality': {'name': 'locality', 'type': str},
            'country': {'name': 'country', 'type': str}},
    'sub': [],
    'finalize': None
}

Circuit = {
    'name': 'Circuit',
    'type': dict,
    'method': _flatten_by_rename,
    'map': {'circuitId': {'name': 'circuitId', 'type': str},
            'url': {'name': 'circuitUrl', 'type': str},
            'circuitName': {'name': 'circuitName', 'type': str}},
    'sub': [Location],
    'finalize': None
}

QualifyingResults = {
    'name': 'QualifyingResults',
    'type': list,
    'method': _flatten_by_rename,
    'map': {'number': {'name': 'number', 'type': save_int},
            'position': {'name': 'position', 'type': save_int},
            'Q1': {'name': 'Q1', 'type': timedelta_from_ergast},
            'Q2': {'name': 'Q2', 'type': timedelta_from_ergast},
            'Q3': {'name': 'Q3', 'type': timedelta_from_ergast}},
    'sub': [Driver, Constructor],
    'finalize': None
}

RaceResults = {
    'name': 'Results',
    'type': list,
    'method': _flatten_by_rename,
    'map': {'number': {'name': 'number', 'type': save_int},
            'position': {'name': 'position', 'type': save_int},
            'positionText': {'name': 'positionText', 'type': str},
            'points': {'name': 'points', 'type': save_float},
            'grid': {'name': 'grid', 'type': save_int},
            'laps': {'name': 'laps', 'type': save_int},
            'status': {'name': 'status', 'type': str}},
    'sub': [Driver, Constructor, TotalRaceTime, FastestLap,
            FastestLapAvgSpeed],
    'finalize': None
}

SprintResults = {
    **RaceResults,  # generate from _RaceResults
    'name': 'SprintResults'
}

DriverStandings = {
    'name': 'DriverStandings',
    'type': list,
    'method': _flatten_by_rename,
    'map': {'position': {'name': 'position', 'type': save_int},
            'positionText': {'name': 'positionText', 'type': str},
            'points': {'name': 'points', 'type': save_float},
            'wins': {'name': 'wins', 'type': save_int}},
    'sub': [Driver, ConstructorsInline],
    'finalize': None
}

ConstructorStandings = {
    **DriverStandings,  # generate from _DriverStandings
    'name': 'ConstructorStandings',
    'sub': [Constructor]
}

Timings = {
    'name': 'Timings',
    'type': list,
    'method': _flatten_inline_list_of_dicts,
    'map': {'driverId': {'name': 'driverId', 'type': str},
            'position': {'name': 'position', 'type': save_int},
            'time': {'name': 'time', 'type': timedelta_from_ergast},
            },
    'sub': [],
    'finalize': None
}

Laps = {
    'name': 'Laps',
    'type': list,
    'method': _lap_timings_flatten_by_rename,
    'map': {'number': {'name': 'number', 'type': save_int}},
    'sub': [Timings],
    'finalize': _merge_dicts_of_lists
}

PitStops = {
    'name': 'PitStops',
    'type': list,
    'method': _flatten_by_rename,
    'map': {'driverId': {'name': 'driverId', 'type': str},
            'stop': {'name': 'stop', 'type': save_int},
            'lap': {'name': 'lap', 'type': save_int},
            'time': {'name': 'time', 'type': time_from_ergast},
            'duration': {'name': 'duration', 'type': timedelta_from_ergast}},
    'sub': [Driver, ConstructorsInline],
    'finalize': None
}

# ##############################
# ### response root elements ###

Seasons = {
    'name': 'Seasons',
    'type': list,
    'method': _flatten_by_rename,
    'map': {'season': {'name': 'season', 'type': save_int},
            'url': {'name': 'seasonUrl', 'type': str}},
    'sub': [],
    'finalize': None
}

__StandingsLists = {
    'name': 'StandingsLists',
    'type': list,
    'method': _flatten_by_rename,
    'map': {'season': {'name': 'season', 'type': save_int},
            'round': {'name': 'round', 'type': save_int}},
    'finalize': None
}

StandingsLists_Driver = {
    **__StandingsLists,
    'sub': [DriverStandings]
}

StandingsLists_Constructor = {
    **__StandingsLists,
    'sub': [ConstructorStandings]
}

__Races = {
    # template for all 'Races' based categories
    'name': 'Races',
    'type': list,
    'method': _flatten_by_rename,
    'map': {'season': {'name': 'season', 'type': save_int},
            'round': {'name': 'round', 'type': save_int},
            'url': {'name': 'raceUrl', 'type': str},
            'raceName': {'name': 'raceName', 'type': str},
            'date': {'name': 'raceDate', 'type': date_from_ergast},
            'time': {'name': 'raceTime', 'type': time_from_ergast}},
    'finalize': None
}

Races_Schedule = {
    **__Races,
    'sub': [Circuit, FirstPractice, SecondPractice, ThirdPractice,
            Qualifying, Sprint]
}

Races_RaceResults = {
    **__Races,
    'sub': [Circuit, RaceResults],
}

Races_QualifyingResults = {
    **__Races,
    'sub': [Circuit, QualifyingResults]
}

Races_SprintResults = {
    **__Races,
    'sub': [Circuit, SprintResults]
}

Races_Laps = {
    **__Races,
    'sub': [Circuit, Laps]
}

Races_PitStops = {
    **__Races,
    'sub': [Circuit, PitStops]
}

Drivers = {
    **Driver,  # from Driver
    'type': list,
    'name': 'Drivers'
}

Constructors = {
    **Constructor,
    'type': list,
    'name': 'Constructors',
}

Circuits = {
    **Circuit,  # from Circuit
    'type': list,
    'name': 'Circuits'
}

Status = {
    'name': 'Status',
    'type': list,
    'method': _flatten_by_rename,
    'map': {'statusId': {'name': 'statusId', 'type': save_int},
            'count': {'name': 'count', 'type': save_int},
            'status': {'name': 'status', 'type': str}},
    'sub': [],
    'finalize': None
}



"""This is a collection of various functions."""
import datetime
import warnings
from functools import reduce
from typing import (
    Dict,
    Optional,
    Tuple,
    Union
)

import numpy as np
import pandas as pd

import fastf1
from fastf1.logger import get_logger


_logger = get_logger(__name__)


def delta_time(
        reference_lap: "fastf1.core.Lap",
        compare_lap: "fastf1.core.Lap"
) -> Tuple[pd.Series, "fastf1.core.Telemetry", "fastf1.core.Telemetry"]:
    """Calculates the delta time of a given lap, along the 'Distance' axis
    of the reference lap.

    .. deprecated:: 3.0.0

    .. warning:: This function should no longer be considered as a stable part
        of the API. Due to the reasons given below, this function will be
        modified or removed at a future point.

    .. warning:: This is a nice gimmick but not actually very accurate which
        is an inherent problem from the way this is calculated currently
        (There may not be a better way though). In comparison with the sector
        times and the differences that can be calculated from these, there are
        notable differences! You should always verify the result against
        sector time differences or find a different way for verification.

    Here is an example that compares the quickest laps of Leclerc and
    Hamilton from Bahrain 2021 Qualifying:

    .. plot::
        :include-source:

        import fastf1 as ff1
        from fastf1 import plotting
        from fastf1 import utils
        from matplotlib import pyplot as plt

        plotting.setup_mpl(misc_mpl_mods=False, color_scheme='fastf1')

        session = ff1.get_session(2021, 'Emilia Romagna', 'Q')
        session.load()
        lec = session.laps.pick_driver('LEC').pick_fastest()
        ham = session.laps.pick_driver('HAM').pick_fastest()

        delta_time, ref_tel, compare_tel = utils.delta_time(ham, lec)
        # ham is reference, lec is compared

        fig, ax = plt.subplots()
        # use telemetry returned by .delta_time for best accuracy,
        # this ensures the same applied interpolation and resampling
        ax.plot(ref_tel['Distance'], ref_tel['Speed'],
                color=plotting.get_team_color(ham['Team'], session))
        ax.plot(compare_tel['Distance'], compare_tel['Speed'],
                color=plotting.get_team_color(lec['Team'], session))

        twin = ax.twinx()
        twin.plot(ref_tel['Distance'], delta_time, '--', color='white')
        twin.set_ylabel("<-- Lec ahead | Ham ahead -->")
        plt.show()

    Args:
        reference_lap: The lap taken as reference
        compare_lap: The lap to compare

    Returns:
        A tuple containing

        - pd.Series of type `float64` with the delta in seconds.
        - :class:`~fastf1.core.Telemetry` for the reference lap
        - :class:`~fastf1.core.Telemetry` for the comparison lap

        Use the return telemetry for plotting to make sure you have
        telemetry data that was created with the same interpolation and
        resampling options!
    """
    warnings.warn("`utils.delta_time` is considered deprecated and will"
                  "be modified or removed in a future release because it has"
                  "a tendency to give inaccurate results.",
                  FutureWarning)

    ref = reference_lap.get_car_data(interpolate_edges=True).add_distance()
    comp = compare_lap.get_car_data(interpolate_edges=True).add_distance()

    def mini_pro(stream):
        # Ensure that all samples are interpolated
        dstream_start = stream[1] - stream[0]
        dstream_end = stream[-1] - stream[-2]
        return np.concatenate(
            [[stream[0] - dstream_start], stream, [stream[-1] + dstream_end]]
        )

    ltime = mini_pro(comp['Time'].dt.total_seconds().to_numpy())
    multiplier = ref.Distance.iat[-1]/comp.Distance.iat[-1]
    ldistance = mini_pro(comp['Distance'].to_numpy())*multiplier
    lap_time = np.interp(ref['Distance'], ldistance, ltime)

    delta = lap_time - ref['Time'].dt.total_seconds()

    return delta, ref, comp


def recursive_dict_get(d: Dict, *keys: str, default_none: bool = False):
    """Recursive dict get. Can take an arbitrary number of keys and returns an
    empty dict if any key does not exist.
    https://stackoverflow.com/a/28225747"""
    ret = reduce(lambda c, k: c.get(k, {}), keys, d)
    if default_none and ret == {}:
        return None
    else:
        return ret


def to_timedelta(x: Union[str, datetime.timedelta]) \
        -> Optional[datetime.timedelta]:
    """Fast timedelta object creation from a time string

    Permissible string formats:

        For example: `13:24:46.320215` with:

            - optional hours and minutes
            - optional microseconds and milliseconds with
              arbitrary precision (1 to 6 digits)

        Examples of valid formats:

            - `24.3564` (seconds + milli/microseconds)
            - `36:54` (minutes + seconds)
            - `8:45:46` (hours, minutes, seconds)

    Args:
        x: timestamp
    """
    # this is faster than using pd.timedelta on a string
    if isinstance(x, str) and len(x):
        try:
            hours, minutes = 0, 0
            if len(hms := x.split(':')) == 3:
                hours, minutes, seconds = hms
            elif len(hms) == 2:
                minutes, seconds = hms
            else:
                seconds = hms[0]

            if '.' in seconds:
                seconds, msus = seconds.split('.')
                if len(msus) < 6:
                    msus = msus + '0' * (6 - len(msus))
                elif len(msus) > 6:
                    msus = msus[0:6]
            else:
                msus = 0

            return datetime.timedelta(
                hours=int(hours), minutes=int(minutes),
                seconds=int(seconds), microseconds=int(msus)
            )

        except Exception as exc:
            _logger.debug(f"Failed to parse timedelta string '{x}'",
                          exc_info=exc)
            return None

    elif isinstance(x, datetime.timedelta):
        return x

    else:
        return None


def to_datetime(x: Union[str, datetime.datetime]) \
        -> Optional[datetime.datetime]:
    """Fast datetime object creation from a date string.

    Permissible string formats:

        For example '2020-12-13T13:27:15.320000Z' with:

            - optional milliseconds and microseconds with
              arbitrary precision (1 to 6 digits)
            - with optional trailing letter 'Z'

        Examples of valid formats:

            - `2020-12-13T13:27:15.320000`
            - `2020-12-13T13:27:15.32Z`
            - `2020-12-13T13:27:15`

    Args:
        x: timestamp
    """
    if isinstance(x, str) and x:
        try:
            date, time = x.strip('Z').split('T')
            year, month, day = date.split('-')
            hours, minutes, seconds = time.split(':')
            if '.' in seconds:
                seconds, msus = seconds.split('.')
                if len(msus) < 6:
                    msus = msus + '0' * (6 - len(msus))
                elif len(msus) > 6:
                    msus = msus[0:6]
            else:
                msus = 0

            return datetime.datetime(
                int(year), int(month), int(day), int(hours),
                int(minutes), int(seconds), int(msus)
            )

        except Exception as exc:
            _logger.debug(f"Failed to parse datetime string '{x}'",
                          exc_info=exc)
            return None

    elif isinstance(x, datetime.datetime):
        return x

    else:
        return None



#!/usr/bin/python
# -*- coding: utf-8 -*-

class Hub:
    def __init__(self, name, connection):
        self.name = name
        self.server = HubServer(name, connection, self)
        self.client = HubClient(name, connection)


class HubServer:
    def __init__(self, name, connection, hub):
        self.name = name
        self.__connection = connection
        self.__hub = hub

    def invoke(self, method, *data):
        message = {
            'H': self.name,
            'M': method,
            'A': data,
            'I': self.__connection.increment_send_counter()
        }
        self.__connection.send(message)


class HubClient(object):
    def __init__(self, name, connection):
        self.name = name
        self.__handlers = {}

        async def handle(**data):
            messages = data['M'] if 'M' in data and len(data['M']) > 0 else {}
            for inner_data in messages:
                hub = inner_data['H'] if 'H' in inner_data else ''
                if hub.lower() == self.name.lower():
                    method = inner_data['M']
                    message = inner_data['A']
                    await self.__handlers[method](message)

        connection.received += handle

    def on(self, method, handler):
        if method not in self.__handlers:
            self.__handlers[method] = handler

    def off(self, method, handler):
        if method in self.__handlers:
            self.__handlers[method] -= handler



import copy
import json
from typing import (
    List,
    Literal,
    Optional,
    Type,
    Union
)

import fastf1.ergast.structure as API
from fastf1 import __version_short__
from fastf1.internals.pandas_base import (
    BaseDataFrame,
    BaseSeries
)
from fastf1.req import Cache


BASE_URL = 'https://ergast.com/api/f1'
HEADERS = {'User-Agent': f'FastF1/{__version_short__}'}


class ErgastResponseMixin:
    """A Mixin class that adds support for pagination to Ergast response
    objects.

    All Ergast response objects provide the methods that are implemented by
    this Mixin.
    """
    _internal_names = ['_response_headers', '_query_filters',
                       '_query_metadata', '_selectors']

    def __init__(self, *args, response_headers: dict, query_filters: dict,
                 metadata: dict, selectors: dict, **kwargs):
        super().__init__(*args, **kwargs)
        self._response_headers = response_headers
        self._query_filters = query_filters
        self._query_metadata = metadata
        self._selectors = selectors

    @property
    def _ergast_constructor(self) -> object:
        return Ergast

    @property
    def total_results(self) -> int:
        """Returns the total number of available results for the request
        associated with this response."""
        return int(self._response_headers.get("total", 0))

    @property
    def is_complete(self) -> bool:
        """Indicates if the response contains all available results for the
        request that is associated with it."""
        # if offset is non-zero, data is missing at the beginning
        if int(self._response_headers.get('offset', 0)) != 0:
            return False

        # can only be complete if limit >= total
        return (int(self._response_headers.get("limit", 0))
                >= int(self._response_headers.get("total", 0)))

    def get_next_result_page(self) -> Union['ErgastSimpleResponse',
                                            'ErgastMultiResponse',
                                            'ErgastRawResponse']:
        """Returns the next page of results within the limit that was specified
        in the request that is associated with this response.

        Raises:
            ValueError: there is no result page after the current page
        """
        n_last = (int(self._response_headers.get("offset", 0))
                  + int(self._response_headers.get("limit", 0)))

        if n_last >= int(self._response_headers.get("total", 0)):
            raise ValueError("No more data after this response.")

        return self._ergast_constructor()._build_default_result(  # noqa: access to builtin
            **self._query_metadata,
            selectors=self._selectors,
            limit=int(self._response_headers.get("limit")),
            offset=n_last
        )

    def get_prev_result_page(self) -> Union['ErgastSimpleResponse',
                                            'ErgastMultiResponse',
                                            'ErgastRawResponse']:
        """Returns the previous page of results within the limit that was
        specified in the request that is associated with this response.

        Raises:
            ValueError: there is no result page before the current page
        """
        n_first = int(self._response_headers.get("offset", 0))

        if n_first <= 0:
            raise ValueError("No more data before this response.")

        limit = int(self._response_headers.get("limit", 0))
        new_offset = max(n_first - limit, 0)

        return self._ergast_constructor()._build_default_result(  # noqa: access to builtin
            **self._query_metadata,
            selectors=self._selectors,
            limit=int(self._response_headers.get("limit")),
            offset=new_offset
        )


class ErgastResultFrame(BaseDataFrame):
    """
    Wraps a Pandas ``DataFrame``. Additionally, this class can be
    initialized from Ergast response data with automatic flattening and type
    casting of the data.

    Args:
        data: Passed through to the DataFrame constructor (must be None if
            `response` is provided)
        category: Reference to a category from ``fastf1.ergast.structure``
            that describes the result data
        response: JSON-like response data from Ergast; used to generate `data`
            from an Ergast response (must be None if `data` is provided)
        auto_cast: Determines if values are automatically cast to the most
            appropriate data type from their original string representation
    """
    _internal_names = BaseDataFrame._internal_names + ['base_class_view']
    _internal_names_set = set(_internal_names)

    def __init__(self, data=None, *,
                 category: Optional[dict] = None,
                 response: Optional[list] = None,
                 auto_cast: bool = True,
                 **kwargs):
        if (data is not None) and (response is not None):
            raise ValueError(f"Cannot initialize {type(self)} with `data`"
                             f"and `response`.")
        if (data is None) and (response is not None):
            data = self._prepare_response(response, category, auto_cast)
        super().__init__(data, **kwargs)

    @classmethod
    def _prepare_response(cls, response: list, category: dict, cast: bool):
        data = copy.deepcopy(response)  # TODO: efficiency?
        for i in range(len(data)):
            _, data[i] = cls._flatten_element(data[i], category, cast)

        if (finalizer := category.get('finalize')) is not None:
            data = finalizer(data)

        return data

    @classmethod
    def _flatten_element(cls, nested: dict, category: dict, cast: bool):
        flat = {}

        # call the categories associated flattening method on the data
        # (operations on 'nested' and 'flat' are inplace, therefore no return)
        category['method'](nested, category, flat, cast=cast)

        # recursively step into subcategories; updated the flattened result
        # dict with the result from the renaming of the subcategory values
        for subcategory in category['sub']:
            if (subname := subcategory['name']) not in nested:
                continue
            _, subflat = cls._flatten_element(
                nested[subname], subcategory, cast
            )
            flat.update(subflat)

        return nested, flat

    @property
    def _constructor_sliced_horizontal(self):
        return ErgastResultSeries


class ErgastResultSeries(BaseSeries):
    """
    Wraps a Pandas ``Series``.

    Currently, no extra functionality is implemented.
    """
    pass


class ErgastRawResponse(ErgastResponseMixin, list):
    """
    Provides the raw JSON-like response data from Ergast.

    This class wraps a ``list`` and adds response information
    and paging (see :class:`ErgastResponseMixin`).

    This "raw" response does not actually contain the complete JSON response
    that the API provides. Instead, only the actual data part of the response
    is returned while metadata (version, query parameters, response length,
    ...) are not included. But metadata is used internally to provide
    pagination and information that are implemented by the
    :class:`ErgastResponseMixin`.

    Args:
        category: Reference to a category from ``fastf1.ergast.structure``
            that describes the result data
        auto_cast: Determines if values are automatically cast to the most
            appropriate data type from their original string representation
    """

    def __init__(self, *, query_result, category, auto_cast, **kwargs):
        if auto_cast:
            query_result = self._prepare_response(query_result, category)

        super().__init__(query_result, **kwargs)

    @classmethod
    def _prepare_response(cls, query_result, category):
        # query_result is a list of json-like data. Each element in that list
        # has the same structure. Iterate over all elements and call the
        # recursive _auto_cast method to convert data types
        query_result = copy.deepcopy(query_result)  # TODO: efficiency?
        for i in range(len(query_result)):
            query_result[i] = cls._auto_cast(query_result[i], category)
        return query_result

    @classmethod
    def _auto_cast(cls, data, category):
        # convert datatypes for all known elements
        for name, mapping in category['map'].items():
            if name not in data:
                continue
            data[name] = mapping['type'](data[name])

        # recursively step into known subcategories and convert data types
        for subcategory in category['sub']:
            if (subname := subcategory['name']) not in data:
                continue
            subcast = cls._auto_cast(data[subname], subcategory)
            data[subname] = subcast

        return data


class ErgastSimpleResponse(ErgastResponseMixin, ErgastResultFrame):
    """
    Provides simple Ergast result data in the form of a Pandas ``DataFrame``.

    This class wraps an :class:`ErgastResultFrame` and adds response
    information and paging (see :class:`ErgastResponseMixin`).
    """
    _internal_names = \
        ErgastResultFrame._internal_names \
        + ErgastResponseMixin._internal_names
    _internal_names_set = set(_internal_names)

    @property
    def _constructor(self) -> Type["ErgastResultFrame"]:
        # drop from ErgastSimpleResponse to ErgastResultFrame, removing the
        # ErgastResponseMixin because a slice of the data is no longer a full
        # response and pagination, ... is therefore not supported anymore
        return ErgastResultFrame


class ErgastMultiResponse(ErgastResponseMixin):
    """
    Provides complex Ergast result data in the form of multiple Pandas
    ``DataFrames``.

    This class additionally offers response information and paging
    (see :class:`ErgastResponseMixin`).

    Note: This object is usually not instantiated by the user. Instead,
    you should use one of the API endpoint methods that are provided by
    :class:`Ergast` get data from the API.

    Example:

    .. doctest::

        >>> from fastf1.ergast import Ergast
        >>> ergast = Ergast(result_type='pandas', auto_cast=True)
        >>> result = ergast.get_race_results(season=2022)

        # The description shows that the result includes data from two
        # grand prix.
        >>> result.description
           season  round  ... locality       country
        0    2022      1  ...   Sakhir       Bahrain
        1    2022      2  ...   Jeddah  Saudi Arabia
        <BLANKLINE>
        [2 rows x 13 columns]

        # As expected, ``result.content`` contains two elements, one for each
        # row of the description
        >>> len(result.content)
        2

        # The first element contains all results from the first of the two
        # grand prix.
        >>> result.content[0]
            number  position  ... fastestLapAvgSpeedUnits  fastestLapAvgSpeed
        0       16         1  ...                     kph             206.018
        1       55         2  ...                     kph             203.501
        2       44         3  ...                     kph             202.469
        ...
        17      11        18  ...                     kph             202.762
        18       1        19  ...                     kph             204.140
        19      10        20  ...                     kph             200.189
        <BLANKLINE>
        [20 rows x 26 columns]

        # The second element is incomplete and only contains the first 10
        # positions of the second Grand Prix. This is because by default,
        # every query on Ergast is limited to 30 result values. You can
        # manually change this limit for each request though.
        >>> result.content[1]
           number  position  ... fastestLapAvgSpeedUnits  fastestLapAvgSpeed
        0       1         1  ...                     kph             242.191
        1      16         2  ...                     kph             242.556
        2      55         3  ...                     kph             241.841
        ...
        7      10         8  ...                     kph             237.796
        8      20         9  ...                     kph             239.562
        9      44        10  ...                     kph             239.001
        <BLANKLINE>
        [10 rows x 26 columns]

    Args:
        response_description: Ergast response containing only the "descriptive"
            information (only data that is available in :attr:`.description`)
        response_data: A list of the "content" data that has been split from
            the Ergast response (data that is available in :attr:`.content`)
        category: A category object from ``fastf1.ergast.structure``
            that defines the main category.
        subcategory: A category object from ``fastf1.ergast.structure``
            that defines the subcategory which is the content data.
        auto_cast: Flag that enables or disables automatic casting from the
            original string representation to the most suitable data type.
    """

    def __init__(self, *args,
                 response_description: dict,
                 response_data: list,
                 category: dict,
                 subcategory: dict,
                 auto_cast: bool,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._description = ErgastResultFrame(response=response_description,
                                              category=category,
                                              auto_cast=auto_cast)
        self._content = [ErgastResultFrame(response=elem,
                                           category=subcategory,
                                           auto_cast=auto_cast)
                         for elem in response_data]

    @property
    def description(self) -> ErgastResultFrame:
        """An :class:`ErgastResultFrame` that describes the data in
        :attr:`.content`.

        Each row of this :class:`ErgastResultFrame` contains the descriptive
        information for one element in :attr:`.content`.
        """
        return self._description

    @property
    def content(self) -> List[ErgastResultFrame]:
        """A ``list`` of :class:`ErgastResultFrame` that contain the main
        response data.

        Descriptive data for each :class:`ErgastResultFrame` is given in the
        corresponding row of :attr:`.description`.
        """
        return self._content


class Ergast:
    """
    The main object that acts as an interface to the Ergast API.

    For each API endpoint, there is a separate method implemented to
    request data.

    Args:
        result_type: Determines the default type of the returned result object

            - 'raw': :class:`~interface.ErgastRawResponse`
            - 'pandas': :class:`~interface.ErgastSimpleResponse` or
              :class:`~interface.ErgastMultiResponse` depending on endpoint

        auto_cast: Determines whether result values are cast from there default
            string representation to a better matching type

        limit: The maximum number of results returned by the API. Defaults to
            30 if not set. Maximum: 1000. See also "Response Paging" on
            https://ergast.com/mrd/.
    """

    def __init__(self,
                 result_type: Literal['raw', 'pandas'] = 'pandas',
                 auto_cast: bool = True,
                 limit: Optional[int] = None):
        self._default_result_type = result_type
        self._default_auto_cast = auto_cast
        self._limit = limit

    @staticmethod
    def _build_url(
            endpoint: str,
            season: Union[Literal['current'], int] = None,
            round: Union[Literal['last'], int] = None,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            lap_number: Optional[int] = None,
            stop_number: Optional[int] = None,
            standings_position: Optional[int] = None
    ) -> str:
        selectors = list()

        if season is not None:
            selectors.append(f"/{season}")
        if round is not None:
            selectors.append(f"/{round}")
        if grid_position is not None:
            selectors.append(f"/grid/{grid_position}")
        if fastest_rank is not None:
            selectors.append(f"/fastest/{fastest_rank}")

        # some special cases: the endpoint may also be used as selector
        # therefore, if the specifier is defined, do not add the endpoint
        # string additionally
        if driver is not None:
            if endpoint == 'drivers':
                endpoint = f"drivers/{driver}"
            else:
                selectors.append(f"/drivers/{driver}")

        if constructor is not None:
            if endpoint == 'constructors':
                endpoint = f"constructors/{constructor}"
            else:
                selectors.append(f"/constructors/{constructor}")

        if circuit is not None:
            if endpoint == 'circuits':
                endpoint = f"circuits/{circuit}"
            else:
                selectors.append(f"/circuits/{circuit}")

        if status is not None:
            if endpoint == 'status':
                endpoint = f"status/{status}"
            else:
                selectors.append(f"/status/{status}")

        if standings_position is not None:
            if endpoint == 'driverStandings':
                endpoint = f"driverStandings/{standings_position}"
            elif endpoint == 'constructorStandings':
                endpoint = f"constructorStandings/{standings_position}"

        if results_position is not None:
            if endpoint in ('results', 'qualifying', 'sprint'):
                endpoint = f"{endpoint}/{results_position}"
            else:
                selectors.append(f"/results/{results_position}")

        if lap_number is not None:
            if endpoint == 'laps':
                endpoint = f"laps/{lap_number}"
            else:
                selectors.append(f"/laps/{lap_number}")

        if stop_number is not None:
            if endpoint == 'pitstops':
                endpoint = f"pitstops/{stop_number}"
            else:
                selectors.append(f"/pitstops/{stop_number}")

        if endpoint is not None:
            selectors.append(f"/{endpoint}")

        return BASE_URL + "".join(selectors) + ".json"

    @classmethod
    def _get(cls, url: str, params: dict) -> Union[dict, list]:
        # request data from ergast and load the returned json data.
        r = Cache.requests_get(url, headers=HEADERS, params=params)
        if r.status_code == 200:
            try:
                return json.loads(r.content.decode('utf-8'))
            except Exception as exc:
                Cache.delete_response(url)  # don't keep a corrupted response
                raise ErgastJsonError(
                    f"Failed to parse Ergast response ({url})"
                ) from exc
        else:
            raise ErgastInvalidRequestError(
                f"Invalid request to Ergast ({url})\n"
                f"Server response: '{r.reason}'"
            )

    @classmethod
    def _build_result(
            cls, *,
            endpoint: str,
            table: str,
            category: dict,
            subcategory: Optional[dict],
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            selectors: Optional[dict] = None,
    ) -> Union[ErgastSimpleResponse,
               ErgastMultiResponse,
               ErgastRawResponse]:
        # query the Ergast database and
        # split the raw response into multiple parts, depending also on what
        # type was selected for the response data format.

        url = cls._build_url(endpoint, **selectors)
        params = {'limit': limit, 'offset': offset}

        # get response and split it into individual parts
        resp = cls._get(url, params)
        resp = resp['MRData']
        body = resp.pop(table)
        # response headers remain in response
        query_result = body.pop(category['name'])
        # query filters remain in body

        query_metadata = {'endpoint': endpoint, 'table': table,
                          'category': category, 'subcategory': subcategory,
                          'result_type': result_type, 'auto_cast': auto_cast}

        if result_type == 'raw':
            return ErgastRawResponse(
                response_headers=resp, query_filters=body,
                metadata=query_metadata, selectors=selectors,
                query_result=query_result, category=category,
                auto_cast=auto_cast
            )

        if result_type == 'pandas':
            # result element description remains in query result
            result_element_data = list()
            if subcategory is not None:
                for i in range(len(query_result)):
                    result_element_data.append(
                        query_result[i].pop(subcategory['name'])
                    )
                return ErgastMultiResponse(
                    response_headers=resp, query_filters=body,
                    metadata=query_metadata, selectors=selectors,
                    response_description=query_result,
                    response_data=result_element_data,
                    category=category, subcategory=subcategory,
                    auto_cast=auto_cast
                )
            else:
                return ErgastSimpleResponse(
                    response_headers=resp, query_filters=body,
                    metadata=query_metadata, selectors=selectors,
                    response=query_result, category=category,
                    auto_cast=auto_cast
                )

    def _build_default_result(
            self, *,
            selectors: dict,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            **kwargs
    ) -> Union[ErgastSimpleResponse,
               ErgastMultiResponse,
               ErgastRawResponse]:
        # use defaults or per-call overrides if specified
        if result_type is None:
            result_type = self._default_result_type
        if auto_cast is None:
            auto_cast = self._default_auto_cast
        if limit is None:
            limit = self._limit

        return self._build_result(
            result_type=result_type,
            auto_cast=auto_cast,
            limit=limit,
            selectors=selectors,
            **kwargs
        )

    # ### endpoints with single-result responses ###
    #
    # can be represented by a DataFrame-like object
    def get_seasons(
            self,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastSimpleResponse, ErgastRawResponse]:
        """Get a list of seasons.

        See: https://ergast.com/mrd/methods/seasons/

        .. ergast-api-map:: Seasons
            :describe-dataframe:

        Args:
            circuit: select a circuit by its circuit id (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            fastest_rank: select fastest by rank number (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastSimpleResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'circuit': circuit,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'fastest_rank': fastest_rank,
                     'status': status}

        return self._build_default_result(endpoint='seasons',
                                          table='SeasonTable',
                                          category=API.Seasons,
                                          subcategory=None,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_race_schedule(
            self,
            season: Union[Literal['current'], int],
            round: Optional[Union[Literal['last'], int]] = None,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastSimpleResponse, ErgastRawResponse]:
        """Get a list of races.

        See: https://ergast.com/mrd/methods/schedule/

        .. ergast-api-map:: Races_Schedule
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            circuit: select a circuit by its circuit id (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            fastest_rank: select fastest by rank number (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastSimpleResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'circuit': circuit,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'fastest_rank': fastest_rank,
                     'status': status}

        return self._build_default_result(endpoint='races',
                                          table='RaceTable',
                                          category=API.Races_Schedule,
                                          subcategory=None,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_driver_info(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastSimpleResponse, ErgastRawResponse]:
        """Get a list of drivers.

        See: https://ergast.com/mrd/methods/drivers/

        .. ergast-api-map:: Drivers
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            circuit: select a circuit by its circuit id (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            fastest_rank: select fastest by rank number (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastSimpleResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'circuit': circuit,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'fastest_rank': fastest_rank,
                     'status': status}

        return self._build_default_result(endpoint='drivers',
                                          table='DriverTable',
                                          category=API.Drivers,
                                          subcategory=None,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_constructor_info(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastSimpleResponse, ErgastRawResponse]:
        """Get a list of constructors.

        See: https://ergast.com/mrd/methods/constructors/

        .. ergast-api-map:: Constructors
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            circuit: select a circuit by its circuit id (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            fastest_rank: select fastest by rank number (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastSimpleResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'circuit': circuit,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'fastest_rank': fastest_rank,
                     'status': status}

        return self._build_default_result(endpoint="constructors",
                                          table='ConstructorTable',
                                          category=API.Constructors,
                                          subcategory=None,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_circuits(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastSimpleResponse, ErgastRawResponse]:
        """Get a list of circuits.

        See: https://ergast.com/mrd/methods/circuits/

        .. ergast-api-map:: Circuits
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            fastest_rank: select fastest by rank number (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastSimpleResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'fastest_rank': fastest_rank,
                     'status': status}

        return self._build_default_result(endpoint='circuits',
                                          table='CircuitTable',
                                          category=API.Circuits,
                                          subcategory=None,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_finishing_status(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastSimpleResponse, ErgastRawResponse]:
        """Get a list of finishing status codes.

        See: https://ergast.com/mrd/methods/status/

        .. ergast-api-map:: Status
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            circuit: select a circuit by its circuit id (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            fastest_rank: select fastest by rank number (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastSimpleResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'circuit': circuit,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'fastest_rank': fastest_rank,
                     'status': status}

        return self._build_default_result(endpoint='status',
                                          table='StatusTable',
                                          category=API.Status,
                                          subcategory=None,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    # ### endpoint with multi-result responses ###
    #
    # example: qualifying results filtered only by season will yield a
    # result for each weekend
    #
    # needs to be represented by multiple DataFrame-like objects
    def get_race_results(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastMultiResponse, ErgastRawResponse]:
        """Get race results for one or multiple races.

        See: https://ergast.com/mrd/methods/results/

        .. ergast-api-map:: Races_RaceResults
            :subcategory: RaceResults
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            circuit: select a circuit by its circuit id (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            fastest_rank: select fastest by rank number (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastMultiResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'circuit': circuit,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'fastest_rank': fastest_rank,
                     'status': status}

        return self._build_default_result(endpoint='results',
                                          table='RaceTable',
                                          category=API.Races_RaceResults,
                                          subcategory=API.RaceResults,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_qualifying_results(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            fastest_rank: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastMultiResponse, ErgastRawResponse]:
        """Get qualifying results for one or multiple qualifying sessions.

        See: https://ergast.com/mrd/methods/qualifying/

        .. ergast-api-map:: Races_QualifyingResults
            :subcategory: QualifyingResults
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            circuit: select a circuit by its circuit id (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            fastest_rank: select fastest by rank number (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastMultiResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'circuit': circuit,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'fastest_rank': fastest_rank,
                     'status': status}

        return self._build_default_result(endpoint='qualifying',
                                          table='RaceTable',
                                          category=API.Races_QualifyingResults,
                                          subcategory=API.QualifyingResults,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_sprint_results(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            circuit: Optional[str] = None,
            constructor: Optional[str] = None,
            driver: Optional[str] = None,
            grid_position: Optional[int] = None,
            results_position: Optional[int] = None,
            status: Optional[str] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastMultiResponse, ErgastRawResponse]:
        """Get sprint results for one or multiple sprints.

        See: https://ergast.com/mrd/methods/sprint/

        .. ergast-api-map:: Races_SprintResults
            :subcategory: SprintResults
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            circuit: select a circuit by its circuit id (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            driver: select a driver by its driver id (default: all)
            grid_position: select a grid position by its number (default: all)
            results_position: select a finishing result by its position
                (default: all)
            status: select by finishing status (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastMultiResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'circuit': circuit,
                     'constructor': constructor,
                     'driver': driver,
                     'grid_position': grid_position,
                     'results_position': results_position,
                     'status': status}

        return self._build_default_result(endpoint='sprint',
                                          table='RaceTable',
                                          category=API.Races_SprintResults,
                                          subcategory=API.SprintResults,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_driver_standings(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            driver: Optional[str] = None,
            standings_position: Optional[int] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastMultiResponse, ErgastRawResponse]:
        """Get driver standings at specific points of a season.

        See: https://ergast.com/mrd/methods/standings/

        .. ergast-api-map:: StandingsLists_Driver
            :subcategory: DriverStandings
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            driver: select a driver by its driver id (default: all)
            standings_position: select a result by position in the standings
                (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastMultiResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'driver': driver,
                     'standings_position': standings_position}

        return self._build_default_result(endpoint='driverStandings',
                                          table='StandingsTable',
                                          category=API.StandingsLists_Driver,
                                          subcategory=API.DriverStandings,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_constructor_standings(
            self,
            season: Optional[Union[Literal['current'], int]] = None,
            round: Optional[Union[Literal['last'], int]] = None,
            constructor: Optional[str] = None,
            standings_position: Optional[int] = None,
            result_type: Optional[Literal['pandas', 'raw']] = None,
            auto_cast: Optional[bool] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None
    ) -> Union[ErgastMultiResponse, ErgastRawResponse]:
        """Get constructor standings at specific points of a season.

        See: https://ergast.com/mrd/methods/standings/

        .. ergast-api-map:: StandingsLists_Constructor
            :subcategory: ConstructorStandings
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            constructor: select a constructor by its constructor id
                (default: all)
            standings_position: select a result by position in the standings
                (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastMultiResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'constructor': constructor,
                     'standings_position': standings_position}

        return self._build_default_result(
            endpoint='constructorStandings',
            table='StandingsTable',
            category=API.StandingsLists_Constructor,
            subcategory=API.ConstructorStandings,
            result_type=result_type,
            auto_cast=auto_cast,
            limit=limit,
            offset=offset,
            selectors=selectors
        )

    def get_lap_times(self,
                      season: Union[Literal['current'], int],
                      round: Union[Literal['last'], int],
                      lap_number: Optional[int] = None,
                      driver: Optional[str] = None,
                      result_type: Optional[Literal['pandas', 'raw']] = None,
                      auto_cast: Optional[bool] = None,
                      limit: Optional[int] = None,
                      offset: Optional[int] = None
                      ) -> Union[ErgastMultiResponse, ErgastRawResponse]:
        """Get sprint results for one or multiple sprints.

        See: https://ergast.com/mrd/methods/laps/

        .. ergast-api-map:: Races_Laps
            :subcategory: Laps
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            lap_number: select lap times by a specific lap number
                (default: all)
            driver: select a driver by its driver id (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastMultiResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'driver': driver,
                     'lap_number': lap_number}

        return self._build_default_result(endpoint='laps',
                                          table='RaceTable',
                                          category=API.Races_Laps,
                                          subcategory=API.Laps,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)

    def get_pit_stops(self,
                      season: Union[Literal['current'], int],
                      round: Union[Literal['last'], int],
                      stop_number: Optional[int] = None,
                      lap_number: Optional[int] = None,
                      driver: Optional[str] = None,
                      result_type: Optional[Literal['pandas', 'raw']] = None,
                      auto_cast: Optional[bool] = None,
                      limit: Optional[int] = None,
                      offset: Optional[int] = None
                      ) -> Union[ErgastMultiResponse, ErgastRawResponse]:
        """Get pit stop information for one or multiple sessions.

        See: https://ergast.com/mrd/methods/standings/

        .. ergast-api-map:: Races_PitStops
            :subcategory: PitStops
            :describe-dataframe:

        Args:
            season: select a season by its year (default: all, oldest first)
            round: select a round by its number (default: all)
            lap_number: select pit stops by a specific lap number
                (default: all)
            stop_number: select pit stops by their stop number
                (default: all)
            driver: select a driver by its driver id (default: all)
            result_type: Overwrites the default result type
            auto_cast: Overwrites the default value for ``auto_cast``
            limit: Overwrites the default value for ``limit``
            offset: An offset into the result set for response paging.
                Defaults to 0 if not set. See also "Response Paging",
                https://ergast.com/mrd/.

        Returns:
            :class:`~interface.ErgastMultiResponse` or
            :class:`~interface.ErgastRawResponse`, depending on the
            ``result_type`` parameter
        """
        selectors = {'season': season,
                     'round': round,
                     'driver': driver,
                     'lap_number': lap_number,
                     'stop_number': stop_number}

        return self._build_default_result(endpoint='pitstops',
                                          table='RaceTable',
                                          category=API.Races_PitStops,
                                          subcategory=API.PitStops,
                                          result_type=result_type,
                                          auto_cast=auto_cast,
                                          limit=limit,
                                          offset=offset,
                                          selectors=selectors)


class ErgastError(Exception):
    """Base class for Ergast API errors."""
    pass


class ErgastJsonError(ErgastError):
    """The response that was returned by the server could not be parsed."""
    pass


class ErgastInvalidRequestError(ErgastError):
    """The server rejected the request because it was invalid."""
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# python compatibility for <3.6
try:
    ModuleNotFoundError
except NameError:
    ModuleNotFoundError = ImportError

# -----------------------------------
# Internal Imports
from ._parameters import WebSocketParameters
from ._queue_events import (
    CloseEvent,
    InvokeEvent
)


# -----------------------------------
# External Imports
try:
    from ujson import (
        dumps,
        loads
    )
except ModuleNotFoundError:
    from json import dumps, loads

import asyncio

import websockets


try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ModuleNotFoundError:
    pass


class Transport:
    def __init__(self, connection):
        self._connection = connection
        self._ws_params = None
        self._conn_handler = None
        self.ws_loop = None
        self.invoke_queue = None
        self.ws = None
        self._set_loop_and_queue()

    # ===================================
    # Public Methods

    def start(self):
        self._ws_params = WebSocketParameters(self._connection)
        self._connect()
        if not self.ws_loop.is_running():
            self.ws_loop.run_forever()

    def send(self, message):
        asyncio.Task(self.invoke_queue.put(InvokeEvent(message)), loop=self.ws_loop)

    def close(self):
        asyncio.Task(self.invoke_queue.put(CloseEvent()), loop=self.ws_loop)

    # -----------------------------------
    # Private Methods

    def _set_loop_and_queue(self):
        try:
            self.ws_loop = asyncio.get_event_loop()
        except RuntimeError:
            self.ws_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.ws_loop)
        self.invoke_queue = asyncio.Queue()

    def _connect(self):
        self._conn_handler = asyncio.ensure_future(self._socket(self.ws_loop), loop=self.ws_loop)

    async def _socket(self, loop):
        async with websockets.connect(self._ws_params.socket_url, extra_headers=self._ws_params.headers,
                                      loop=loop) as self.ws:
            self._connection.started = True
            await self._master_handler(self.ws)

    async def _master_handler(self, ws):
        consumer_task = asyncio.ensure_future(self._consumer_handler(ws), loop=self.ws_loop)
        producer_task = asyncio.ensure_future(self._producer_handler(ws), loop=self.ws_loop)
        done, pending = await asyncio.wait([consumer_task, producer_task],
                                           return_when=asyncio.FIRST_EXCEPTION)

        for task in pending:
            task.cancel()

    async def _consumer_handler(self, ws):
        while True:
            message = await ws.recv()
            if len(message) > 0:
                data = loads(message)
                await self._connection.received.fire(**data)

    async def _producer_handler(self, ws):
        while True:
            try:
                event = await self.invoke_queue.get()
                if event is not None:
                    if event.type == 'INVOKE':
                        await ws.send(dumps(event.message))
                    elif event.type == 'CLOSE':
                        await ws.close()
                        while ws.open is True:
                            await asyncio.sleep(0.1)
                        else:
                            self._connection.started = False
                            break
                else:
                    break
                self.invoke_queue.task_done()
            except Exception as e:
                raise e



from typing import Dict

from fastf1.plotting._constants.base import (
    Compounds,
    Team,
    TeamColors
)


Teams: Dict[str, Team] = {
    'alpine': Team(
        ShortName='Alpine',
        TeamColor=TeamColors(
            Official='#0093cc',
            Default='#ff87bc'
        )
    ),
    'aston martin': Team(
        ShortName='Aston Martin',
        TeamColor=TeamColors(
            Official='#229971',
            Default='#00665f'
        )
    ),
    'ferrari': Team(
        ShortName='Ferrari',
        TeamColor=TeamColors(
            Official='#e8002d',
            Default='#e8002d'
        )
    ),
    'haas': Team(
        ShortName='Haas',
        TeamColor=TeamColors(
            Official='#b6babd',
            Default='#b6babd'
        )
    ),
    'mclaren': Team(
        ShortName='McLaren',
        TeamColor=TeamColors(
            Official='#ff8000',
            Default='#ff8000'
        )
    ),
    'mercedes': Team(
        ShortName='Mercedes',
        TeamColor=TeamColors(
            Official='#27f4d2',
            Default='#27f4d2'
        )
    ),
    'rb': Team(
        ShortName='RB',
        TeamColor=TeamColors(
            Official='#6692ff',
            Default='#364aa9'
        )
    ),
    'red bull': Team(
        ShortName='Red Bull',
        TeamColor=TeamColors(
            Official='#3671c6',
            Default='#0600ef'
        )
    ),
    'sauber': Team(
        ShortName='Sauber',
        TeamColor=TeamColors(
            Official='#52e252',
            Default='#00e700'
        )
    ),
    'williams': Team(
        ShortName='Williams',
        TeamColor=TeamColors(
            Official='#64c4ff',
            Default='#00a0dd'
        )
    )
}

# TODO: future proofing?
CompoundColors: Dict[Compounds, str] = {
    Compounds.Soft: "#da291c",
    Compounds.Medium: "#ffd12e",
    Compounds.Hard: "#f0f0ec",
    Compounds.Intermediate: "#43b02a",
    Compounds.Wet: "#0067ad",
    Compounds.Unknown: "#00ffff",
    Compounds.TestUnknown: "#434649"
}



from typing import Dict

from fastf1.plotting._constants.base import (
    Compounds,
    Team,
    TeamColors
)


Teams: Dict[str, Team] = {
    'alfa romeo': Team(
        ShortName='Alfa Romeo',
        TeamColor=TeamColors(
            Official='#C92D4B',
            Default='#900000'
        )
    ),
    'alphatauri': Team(
        ShortName='AlphaTauri',
        TeamColor=TeamColors(
            Official='#5E8FAA',
            Default='#2b4562'
        )
    ),
    'alpine': Team(
        ShortName='Alpine',
        TeamColor=TeamColors(
            Official='#2293D1',
            Default='#fe86bc'
        )
    ),
    'aston martin': Team(
        ShortName='Aston Martin',
        TeamColor=TeamColors(
            Official='#358C75',
            Default='#00665e'
        )
    ),
    'ferrari': Team(
        ShortName='Ferrari',
        TeamColor=TeamColors(
            Official='#F91536',
            Default='#da291c'
        )
    ),
    'haas': Team(
        ShortName='Haas',
        TeamColor=TeamColors(
            Official='#b6babd',
            Default='#b6babd'
        )
    ),
    'mclaren': Team(
        ShortName='McLaren',
        TeamColor=TeamColors(
            Official='#F58020',
            Default='#ff8000'
        )
    ),
    'mercedes': Team(
        ShortName='Mercedes',
        TeamColor=TeamColors(
            Official='#6CD3BF',
            Default='#00f5d0'
        )
    ),
    'red bull': Team(
        ShortName='Red Bull',
        TeamColor=TeamColors(
            Official='#3671C6',
            Default='#0600ef'
        )
    ),
    'williams': Team(
        ShortName='Williams',
        TeamColor=TeamColors(
            Official='#37BEDD',
            Default='#00a0dd'
        )
    )
}

CompoundColors: Dict[Compounds, str] = {
    Compounds.Soft: "#da291c",
    Compounds.Medium: "#ffd12e",
    Compounds.Hard: "#f0f0ec",
    Compounds.Intermediate: "#43b02a",
    Compounds.Wet: "#0067ad",
    Compounds.Unknown: "#00ffff",
    Compounds.TestUnknown: "#434649"
}



from dataclasses import dataclass
from typing import Optional

import numpy as np
import pandas as pd

import fastf1.core
from fastf1.mvapi.api import get_circuit
from fastf1.mvapi.internals import _logger


@dataclass
class CircuitInfo:
    """Holds information about the circuit that is useful for visualizing and
    annotating data.

    ``corners``, ``marshal_lights`` and ``marshal_sectors`` are track markers
    that all use a similar DataFrame format.
    Each set of these track markers has the following DataFrame columns:

        Columns: ``X <float>, Y <float>, Number <int>, Letter <str>,
        Angle <float>, Distance <float>``

        - ``X`` and ``Y`` specify the position on the track map
        - ``Number`` is the number of the corner. ``Letter`` is optionally used
          to differentiate corners with the same number on some circuits,
          e.g. "2A".
        - ``Angle`` is an angle in degrees, used to visually offset the
          marker's placement on a track map in a logical direction (usually
          orthogonal to the track).
        - ``Distance`` is the location of the marker as a distance from the
          start/finish line. This value needs to be computed using car
          telemetry as a reference. It will therefore only be available,
          if telemetry data is loaded.

    .. note:: This data has been manually created and is not highly accurate
        but sufficient for visualization. A big thanks to MultiViewer
        (https://multiviewer.app/) for providing this data to FastF1.
    """

    corners: pd.DataFrame
    """Location of corners.

    (DataFrame format described above)
    """

    marshal_lights: pd.DataFrame
    """Location of marshal lights.

    (DataFrame format described above)
    """

    marshal_sectors: pd.DataFrame
    """Location of marshal sectors.

    (DataFrame format described above)
    """

    rotation: float
    """Rotation of the circuit in degrees. This can be used to rotate the
    coordinate system of the telemetry (position) data to match the orientation
    of the official track map.
    """

    def add_marker_distance(self, reference_lap: "fastf1.core.Lap"):
        """:meta private:
        Computes the 'Distance' value for each track marker using the
        telemetry data of a provided reference lap.

        The distance values are computed using a best-fit approach by selecting
        the distance value of a telemetry sample, where the squared error of
        the xy-coordinates between telemetry data and track marker position
        is minimal."""

        # get merged reference telemetry and limit to data source 'pos' so that
        # only position values are used that are not interpolated
        try:
            tel = reference_lap.get_telemetry(frequency='original')
        except fastf1.core.DataNotLoadedError:
            _logger.warning("Failed to generate marker distance information: "
                            "telemetry data has not been loaded")
            return

        if tel.empty:
            _logger.warning("Failed to generate marker distance information: "
                            "telemetry data is empty")
            return

        tel = tel[tel['Source'] == 'pos']
        # create a numpy array of xy track coordinates
        xy_ref_array = tel.loc[:, ('X', 'Y')].to_numpy()

        for df in (self.corners, self.marshal_sectors, self.marshal_lights):
            # create a numpy array of xy marker arrays
            marker_xy_array = df.loc[:, ('X', 'Y')].to_numpy()

            n_markers = marker_xy_array.shape[0]

            # create an array of xy track coordinates with an additional
            # dimension that has the size of the number of markers
            xy_array = xy_ref_array \
                .reshape((1, -1, 2)) \
                .repeat(n_markers, axis=0)

            # subtract each marker position from a full set of track
            # coordinates and calculate the squared xy error
            diff = xy_array - marker_xy_array.reshape((-1, 1, 2))
            e = diff[..., 0] ** 2 + diff[..., 1] ** 2

            # return the index in the track coordinates at which the squared
            # error is minimal for each marker
            indices = np.nanargmin(e, axis=1)

            # get the best distance value for each marker from the telemetry
            # data and add them to the marker data frame
            distances = tel.iloc[indices]['Distance'].to_list()
            df['Distance'] = distances


def get_circuit_info(*, year: int, circuit_key: int) -> Optional[CircuitInfo]:
    """:meta private:
    Load circuit information from the MultiViewer API and convert it into
    as :class:``SessionInfo`` object.

    Args:
        year: The championship year
        circuit_key: The unique circuit key (defined by the F1 livetiming API)
    """
    data = get_circuit(year=year, circuit_key=circuit_key)

    if not data:
        _logger.warning("Failed to load circuit info")
        return None

    ret = list()
    for cat in ('corners', 'marshalLights', 'marshalSectors'):
        rows = list()
        for entry in data[cat]:
            rows.append((
                float(entry.get('trackPosition', {}).get('x', 0.0)),
                float(entry.get('trackPosition', {}).get('y', 0.0)),
                int(entry.get('number', 0)),
                str(entry.get('letter', "")),
                float(entry.get('angle', 0.0)),
                np.nan
            ))
        ret.append(
            pd.DataFrame(
                rows,
                columns=['X', 'Y', 'Number', 'Letter', 'Angle', 'Distance']
            )
        )

    rotation = float(data.get('rotation', 0.0))

    circuit_info = CircuitInfo(
        corners=ret[0],
        marshal_lights=ret[1],
        marshal_sectors=ret[2],
        rotation=rotation
    )

    return circuit_info



"""Qualifying results overview
==============================

Plot the qualifying result with visualization the fastest times.
"""


import matplotlib.pyplot as plt
import pandas as pd
from timple.timedelta import strftimedelta

import fastf1
import fastf1.plotting
from fastf1.core import Laps


# Enable Matplotlib patches for plotting timedelta values
fastf1.plotting.setup_mpl(mpl_timedelta_support=True, misc_mpl_mods=False,
                          color_scheme=None)


session = fastf1.get_session(2021, 'Spanish Grand Prix', 'Q')
session.load()


##############################################################################
# First, we need to get an array of all drivers.

drivers = pd.unique(session.laps['Driver'])
print(drivers)


##############################################################################
# After that we'll get each driver's fastest lap, create a new laps object
# from these laps, sort them by lap time and have pandas reindex them to
# number them nicely by starting position.

list_fastest_laps = list()
for drv in drivers:
    drvs_fastest_lap = session.laps.pick_driver(drv).pick_fastest()
    list_fastest_laps.append(drvs_fastest_lap)
fastest_laps = Laps(list_fastest_laps) \
    .sort_values(by='LapTime') \
    .reset_index(drop=True)


##############################################################################
# The plot is nicer to look at and more easily understandable if we just plot
# the time differences. Therefore, we subtract the fastest lap time from all
# other lap times.

pole_lap = fastest_laps.pick_fastest()
fastest_laps['LapTimeDelta'] = fastest_laps['LapTime'] - pole_lap['LapTime']


##############################################################################
# We can take a quick look at the laps we have to check if everything
# looks all right. For this, we'll just check the 'Driver', 'LapTime'
# and 'LapTimeDelta' columns.

print(fastest_laps[['Driver', 'LapTime', 'LapTimeDelta']])


##############################################################################
# Finally, we'll create a list of team colors per lap to color our plot.
team_colors = list()
for index, lap in fastest_laps.iterlaps():
    color = fastf1.plotting.get_team_color(lap['Team'], session=session)
    team_colors.append(color)


##############################################################################
# Now, we can plot all the data
fig, ax = plt.subplots()
ax.barh(fastest_laps.index, fastest_laps['LapTimeDelta'],
        color=team_colors, edgecolor='grey')
ax.set_yticks(fastest_laps.index)
ax.set_yticklabels(fastest_laps['Driver'])

# show fastest at the top
ax.invert_yaxis()

# draw vertical lines behind the bars
ax.set_axisbelow(True)
ax.xaxis.grid(True, which='major', linestyle='--', color='black', zorder=-1000)
# sphinx_gallery_defer_figures


##############################################################################
# Finally, give the plot a meaningful title

lap_time_string = strftimedelta(pole_lap['LapTime'], '%m:%s.%ms')

plt.suptitle(f"{session.event['EventName']} {session.event.year} Qualifying\n"
             f"Fastest Lap: {lap_time_string} ({pole_lap['Driver']})")

plt.show()



# flake8: noqa
"""Who can still win the drivers WDC?
======================================

Calculates which drivers still has chance to win the WDC.
Simplified since it doesn't compare positions if points are equal.

This example implements 3 functions that it then uses to calculate
its result.
"""

import fastf1
from fastf1.ergast import Ergast


##############################################################################
# For this example, we are looking at the 2023 season.
# We want to know who can theoretically still win the drivers' championship
# after the first 15 races.

SEASON = 2023
ROUND = 15


##############################################################################
# Get the current driver standings from Ergast.
# Reference https://docs.fastf1.dev/ergast.html#fastf1.ergast.Ergast.get_driver_standings
def get_drivers_standings():
    ergast = Ergast()
    standings = ergast.get_driver_standings(season=SEASON, round=ROUND)
    return standings.content[0]


##############################################################################
# We need a function to calculates the maximum amount of points possible if a
# driver wins everything left of the season.
# https://en.wikipedia.org/wiki/List_of_Formula_One_World_Championship_points_scoring_systems
def calculate_max_points_for_remaining_season():
    POINTS_FOR_SPRINT = 8 + 25 + 1  # Winning the sprint, race and fastest lap
    POINTS_FOR_CONVENTIONAL = 25 + 1  # Winning the race and fastest lap

    events = fastf1.events.get_event_schedule(SEASON, backend='ergast')
    events = events[events['RoundNumber'] > ROUND]
    # Count how many sprints and conventional races are left
    sprint_events = len(events.loc[events["EventFormat"] == "sprint_shootout"])
    conventional_events = len(events.loc[events["EventFormat"] == "conventional"])

    # Calculate points for each
    sprint_points = sprint_events * POINTS_FOR_SPRINT
    conventional_points = conventional_events * POINTS_FOR_CONVENTIONAL

    return sprint_points + conventional_points


##############################################################################
# For each driver we will see if there is a chance to get more points than
# the current leader. We assume the leader gets no more points and the
# driver gets the theoretical maximum amount of points.
#
# We currently don't consider the case of two drivers getting equal points
# since its more complicated and would require comparing positions.
def calculate_who_can_win(driver_standings, max_points):
    LEADER_POINTS = int(driver_standings.loc[0]['points'])

    for i, _ in enumerate(driver_standings.iterrows()):
        driver = driver_standings.loc[i]
        driver_max_points = int(driver["points"]) + max_points
        can_win = 'No' if driver_max_points < LEADER_POINTS else 'Yes'

        print(f"{driver['position']}: {driver['givenName'] + ' ' + driver['familyName']}, "
              f"Current points: {driver['points']}, "
              f"Theoretical max points: {driver_max_points}, "
              f"Can win: {can_win}")


##############################################################################
# Now using the 3 functions above we can use them to calculate who
# can still win.

# Get the current drivers standings
driver_standings = get_drivers_standings()

# Get the maximum amount of points
points = calculate_max_points_for_remaining_season()

# Print which drivers can still win
calculate_who_can_win(driver_standings, points)



"""Plot speed traces with corner annotations
============================================

Plot the speed over the course of a lap and add annotations to mark corners.
"""


import matplotlib.pyplot as plt

import fastf1.plotting


# Enable Matplotlib patches for plotting timedelta values and load
# FastF1's dark color scheme
fastf1.plotting.setup_mpl(mpl_timedelta_support=True, misc_mpl_mods=False,
                          color_scheme='fastf1')

# load a session and its telemetry data
session = fastf1.get_session(2021, 'Spanish Grand Prix', 'Q')
session.load()

##############################################################################
# First, we select the fastest lap and get the car telemetry data for this
# lap.

fastest_lap = session.laps.pick_fastest()
car_data = fastest_lap.get_car_data().add_distance()

##############################################################################
# Next, load the circuit info that includes the information about the location
# of the corners.

circuit_info = session.get_circuit_info()

##############################################################################
# Finally, we create a plot and plot the speed trace as well as the corner
# markers.

team_color = fastf1.plotting.get_team_color(fastest_lap['Team'],
                                            session=session)

fig, ax = plt.subplots()
ax.plot(car_data['Distance'], car_data['Speed'],
        color=team_color, label=fastest_lap['Driver'])

# Draw vertical dotted lines at each corner that range from slightly below the
# minimum speed to slightly above the maximum speed.
v_min = car_data['Speed'].min()
v_max = car_data['Speed'].max()
ax.vlines(x=circuit_info.corners['Distance'], ymin=v_min-20, ymax=v_max+20,
          linestyles='dotted', colors='grey')

# Plot the corner number just below each vertical line.
# For corners that are very close together, the text may overlap. A more
# complicated approach would be necessary to reliably prevent this.
for _, corner in circuit_info.corners.iterrows():
    txt = f"{corner['Number']}{corner['Letter']}"
    ax.text(corner['Distance'], v_min-30, txt,
            va='center_baseline', ha='center', size='small')

ax.set_xlabel('Distance in m')
ax.set_ylabel('Speed in km/h')
ax.legend()

# Manually adjust the y-axis limits to include the corner numbers, because
# Matplotlib does not automatically account for text that was manually added.
ax.set_ylim([v_min - 40, v_max + 20])

plt.show()



"""
All HTTP requests that are performed by FastF1 go through its caching and
rate limiting system.

Caching is enabled by default in FastF1 and most of the time, you do not need
to worry about caching at all. It will simply happen automatically in the
background and speed up your programs. Disabling the cache is highly
discouraged and will generally slow down your programs.

Rate limits are applied at all times. Requests that can be served from the
cache do not count towards any rate limits. Having the cache enabled can
therefore virtually increase the rate limits.

When rate limits are exceeded, FastF1 will either...

- throttle the rate of requests, if small delays are sufficient to stay within
  the limit (soft rate limit)

- raise a :class:`fastf1.RateLimitExceededError` (hard rate limit)

"""

import collections
import datetime
import functools
import math
import os
import pickle
import re
import sys
import time
from typing import (
    Optional,
    Tuple
)

import requests
from requests_cache import CacheMixin

from fastf1.logger import get_logger


_logger = get_logger(__name__)


# A NOTE TO EVERYBODY WHO READS THIS CODE
# ##############################################
# Rate limits are defined for technical reasons.
# They are not created to simply annoy you even if they may feel annoying.
#
# Some of the APIs that FastF1 accesses are provided by individuals, free
# of charge and in their spare time. Because of that, they may have very
# limited server capacity. We should accept that and be grateful that they
# even exist in the first place.
# Other APIs may be provided by larger companies. But consequently they
# also need to cope with more traffic. We should accept their API limits as
# well.
#
# IN SHORT:
# Please do not edit API limits! If you run into API limits, it is more often
# than not the case that your code can be optimized to prevent this.
# Please optimize your code!
#
# Violating the API limits may get you or even the whole FastF1 project blocked
# from accessing a specific API. This has happened before and just causes
# unnecessary hassle for many people.


class _MinIntervalLimitDelay:
    """Ensure that there is at least a minimum delay between each request.

    Sleeps for the remaining amount of time if the last request was more recent
    than allowed by the minimum interval rule.
    """
    def __init__(self, interval: float):
        self._interval: float = interval
        self._t_last: float = 0.0

    def limit(self):
        t_now = time.time()
        if (delta := (t_now - self._t_last)) < self._interval:
            time.sleep(self._interval - delta)
        self._t_last = t_now


class _CallsPerIntervalLimitRaise:
    """Ensures that there is a maximum number of requests within a fixed
    interval of time.

    If the maximum number of allowed requests within this interval is exceeded,
    a :class:`RateLimitExceeded` exception is raised.
    """
    def __init__(self, calls: int, interval: float, info: str):
        self._interval: float = interval
        self._timestamps = collections.deque(maxlen=calls)
        self._info = info

    def limit(self):
        self._timestamps.append(time.time())
        if len(self._timestamps) == self._timestamps.maxlen:
            if self._timestamps[0] > (time.time() - self._interval):
                raise RateLimitExceededError(self._info)


class _SessionWithRateLimiting(requests.Session):
    """Apply rate limiters to requests that match a URL pattern.
    """
    _RATE_LIMITS = {
        # limits on ergast.com
        re.compile(r"^https?://(\w+\.)?ergast\.com.*"): [
            _MinIntervalLimitDelay(0.25),
            # soft limit 4 calls/sec
            _CallsPerIntervalLimitRaise(200, 60*60, "ergast.com: 200 calls/h")
            # hard limit 200 calls/h
        ]
    }

    def send(self, request, **kwargs):
        # patches rate limiting into `requests.send`
        for pattern, limiters in self._RATE_LIMITS.items():
            # match url pattern
            if pattern.match(request.url):
                for lim in limiters:
                    # apply all defined limiters
                    lim.limit()

        return super().send(request, **kwargs)


class _CachedSessionWithRateLimiting(CacheMixin, _SessionWithRateLimiting):
    """Equivalent of ``requests_cache.CachedSession```but using
    :class:`_SessionWithRateLimiting` as base instead of ``requests.Session``.
    """
    pass


class _MetaCache(type):
    def __repr__(self):
        # implements __repr__ for the Cache class itself
        if self._CACHE_DIR:
            path = self._CACHE_DIR
            size = self._convert_size(self._get_size(path))
            return f"FastF1 cache ({size}) {path}"

        return "FastF1 cache - not configured"


class Cache(metaclass=_MetaCache):
    """Pickle and requests based API cache.

    Fast-F1 will per default enable caching. While this can be disabled, it
    should almost always be left enabled to speed up the runtime of your
    0scripts and to prevent exceeding the rate limit of api servers.

    The default cache directory is defined, in order of precedence, in one
    of the following ways:

    #. A call to :func:`enable_cache`
    #. The value of the environment variable ``FASTF1_CACHE``
    #. An OS dependent default cache directory

    See below for more information on default cache directories.

    The following class-level functions are used to set up, enable and
    (temporarily) disable caching.

    .. autosummary::
        fastf1.Cache.enable_cache
        fastf1.Cache.clear_cache
        fastf1.Cache.get_cache_info
        fastf1.Cache.disabled
        fastf1.Cache.set_disabled
        fastf1.Cache.set_enabled
        fastf1.Cache.offline_mode

    The parsed API data will be saved as a pickled object.
    Raw GET and POST requests are cached in a sqlite db using the
    'requests-cache' module.

    Requests that can be served from the cache do not count towards any
    API rate limits.

    The cache has two "stages":

    - Stage 1: Caching of raw GET requests. This works for all requests.
      Cache control is employed to refresh the cached data periodically.
    - Stage 2: Caching of the parsed data. This saves a lot of time when
      running your scripts,  as parsing of the data is computationally
      expensive. Stage 2 caching is only used for some api functions.

    You can explicitly configure right at the beginning of your script:

        >>> import fastf1
        >>> fastf1.Cache.enable_cache('path/to/cache')  # doctest: +SKIP
        # change cache directory to an existing empty directory on your machine
        >>> session = fastf1.get_session(2021, 5, 'Q')
        >>> # ...

    An alternative way to set the cache directory is to configure an
    environment variable `FASTF1_CACHE`. However, this value will be
    ignored if `Cache.enable_cache()` is called.

    If no explicit location is provided, Fast-F1 will use a default location
    depending on operating system.

    - Windows: `%LOCALAPPDATA%\\\\Temp\\\\fastf1`
    - macOS: `~/Library/Caches/fastf1`
    - Linux: `~/.cache/fastf1` if `~/.cache` exists otherwise `~/.fastf1`

    Cached data can be deleted at any time to reclaim disk space. However,
    this also means you will have to redownload the same data again if you
    need which will lead to reduced performance.
    """
    _CACHE_DIR = None
    # version of the api parser code (unrelated to release version number)
    _API_CORE_VERSION = 12
    _IGNORE_VERSION = False
    _FORCE_RENEW = False

    _requests_session_cached: Optional[_CachedSessionWithRateLimiting] = None
    _requests_session: requests.Session = _SessionWithRateLimiting()
    _default_cache_enabled = False  # flag to ensure that warning about disabled cache is logged once only # noqa: E501
    _tmp_disabled = False
    _ci_mode = False

    _request_counter = 0  # count uncached requests for debugging purposes

    @classmethod
    def enable_cache(
            cls, cache_dir: str, ignore_version: bool = False,
            force_renew: bool = False,
            use_requests_cache: bool = True):
        """Enables the API cache.

        Args:
            cache_dir: Path to the directory which should be used to store
                cached data. Path needs to exist.
            ignore_version: Ignore if cached data was created with a different
                version of the API parser (not recommended: this can cause
                crashes or unrecognized errors as incompatible data may be
                loaded)
            force_renew: Ignore existing cached data. Download data and update
                the cache instead.
            use_requests_cache: Do caching of the raw GET and POST requests.
        """
        # Allow users to use paths such as %LOCALAPPDATA%
        cache_dir = os.path.expandvars(cache_dir)

        # Allow users to use paths such as ~user or ~/
        cache_dir = os.path.expanduser(cache_dir)

        if not os.path.exists(cache_dir):
            raise NotADirectoryError("Cache directory does not exist! Please "
                                     "check for typos or create it first.")
        cls._CACHE_DIR = cache_dir
        cls._IGNORE_VERSION = ignore_version
        cls._FORCE_RENEW = force_renew
        if use_requests_cache:
            cls._requests_session_cached = _CachedSessionWithRateLimiting(
                cache_name=os.path.join(cache_dir, 'fastf1_http_cache'),
                backend='sqlite',
                allowable_methods=('GET', 'POST'),
                expire_after=datetime.timedelta(hours=12),
                cache_control=True,
                stale_if_error=True,
                filter_fn=cls._custom_cache_filter
            )
            if force_renew:
                cls._requests_session_cached.cache.clear()

    @classmethod
    def requests_get(cls, *args, **kwargs):
        """Wraps `requests.Session().get()` with caching if enabled.

        All GET requests that require caching should be performed through this
        wrapper. Caching will be done if the module-wide cache has been
        enabled. Else, `requests.Session().get()` will be called without any
        caching.
        """
        cls._enable_default_cache()
        if (cls._requests_session_cached is None) or cls._tmp_disabled:
            cls._request_counter += 1
            return cls._requests_session.get(*args, **kwargs)

        if cls._ci_mode:
            # try to return a cached response first
            resp = cls._requests_session_cached.get(
                *args, only_if_cached=True, **kwargs)
            # 504 indicates that no cached response was found
            if resp.status_code != 504:
                return resp

        cls._request_counter += 1
        return cls._requests_session_cached.get(*args, **kwargs)

    @classmethod
    def requests_post(cls, *args, **kwargs):
        """Wraps `requests.Session().post()` with caching if enabled.

        All POST requests that require caching should be performed through this
        wrapper. Caching will be done if the module-wide cache has been
        enabled. Else, `requests.Session().get()` will be called without any
        caching.
        """
        cls._enable_default_cache()
        if (cls._requests_session_cached is None) or cls._tmp_disabled:
            cls._request_counter += 1
            return cls._requests_session.post(*args, **kwargs)

        if cls._ci_mode:
            # try to return a cached response first
            resp = cls._requests_session_cached.post(
                *args, only_if_cached=True, **kwargs)
            # 504 indicates that no cached response was found
            if resp.status_code != 504:
                return resp

        cls._request_counter += 1
        return cls._requests_session_cached.post(*args, **kwargs)

    @classmethod
    def delete_response(cls, url):
        """Deletes a single cached response from the cache, if caching is
        enabled. If caching is not enabled, this call is ignored."""
        if cls._requests_session_cached is not None:
            cls._requests_session_cached.cache.delete(urls=[url])

    @staticmethod
    def _custom_cache_filter(response: requests.Response):
        # this function provides custom filtering to decide which responses
        # get cached

        # workaround for Ergast returning error with status code 200
        if 'Unable to select database' in response.text:
            return False

        return True

    @classmethod
    def clear_cache(cls, cache_dir=None, deep=False):
        """Clear all cached data.

        Deletes all files in the cache directory. By default, it will clear
        the default cache directory. However, if a cache directory is
        provided as an argument this will be cleared instead. Optionally,
        the requests cache can be cleared too.

        Can be called without enabling the cache first.

        Deleting specific events or sessions is not supported but can be done
        manually (stage 2 cache). The cached data is structured by year, event
        and session. The structure is more or less self-explanatory. To delete
        specific events or sessions delete the corresponding folder within the
        cache directory. Deleting specific requests from the requests cache
        (stage 1) is not possible. To delete the requests cache only, delete
        the sqlite file in the root of the cache directory.

        Args:
            cache_dir (str): Path to the directory which is used to store
                cached data.
            deep (bool): Clear the requests cache (stage 1) too.
        """
        if cache_dir is None:
            if cls._CACHE_DIR is None:
                cache_dir = cls._get_default_cache_path()
            else:
                cache_dir = cls._CACHE_DIR

        # We need to expand the directory to support ~/
        cache_dir = os.path.expandvars(cache_dir)
        cache_dir = os.path.expanduser(cache_dir)
        if not os.path.exists(cache_dir):
            raise NotADirectoryError("Cache directory does not exist!")

        for dirpath, dirnames, filenames in os.walk(cache_dir):
            for filename in filenames:
                if filename.endswith('.ff1pkl'):
                    os.remove(os.path.join(dirpath, filename))

        if deep:
            cache_db_path = os.path.join(cache_dir, 'fastf1_http_cache.sqlite')
            if os.path.exists(cache_db_path):
                os.remove(cache_db_path)

    @classmethod
    def api_request_wrapper(cls, func):
        """Wrapper function for adding stage 2 caching to api functions.

        Args:
            func: function to be wrapped

        Returns:
            The wrapped function
        """

        @functools.wraps(func)
        def _cached_api_request(api_path, **func_kwargs):
            if cls._CACHE_DIR and not cls._tmp_disabled:
                # caching is enabled
                func_name = str(func.__name__)
                cache_file_path = cls._get_cache_file_path(api_path, func_name)

                if os.path.isfile(cache_file_path):
                    if cls._ci_mode:
                        # skip pickle cache in ci mode so that API parser code
                        # is always executed. Only http cache is active
                        return func(api_path, **func_kwargs)

                    # file exists already, try to load it
                    try:
                        cached = pickle.load(open(cache_file_path, 'rb'))
                    except:  # noqa: E722 (bare except)
                        # don't like the bare exception clause but who knows
                        # which dependency will raise which internal exception
                        # after it was updated
                        cached = None

                    if (cached is not None) and cls._data_ok_for_use(cached):
                        # cached data is ok for use, return it
                        _logger.info(f"Using cached data for {func_name}")
                        return cached['data']

                    else:
                        # cached data needs to be downloaded again and updated
                        _logger.info(f"Updating cache for {func_name}...")
                        data = func(api_path, **func_kwargs)

                        if data is not None:
                            cls._write_cache(data, cache_file_path)
                            _logger.info("Cache updated!")
                            return data

                        _logger.critical(
                            "A cache update is required but the data failed "
                            "to download. Cannot continue!\nYou may force to "
                            "ignore a cache version mismatch by using the "
                            "`ignore_version=True` keyword when enabling the "
                            "cache (not recommended)."
                        )
                        exit()

                else:  # cached data does not yet exist for this api request
                    _logger.info(f"No cached data found for {func_name}. "
                                 f"Loading data...")
                    data = func(api_path, **func_kwargs)
                    if data is not None:
                        cls._write_cache(data, cache_file_path)
                        _logger.info("Data has been written to cache!")
                        return data

                    _logger.critical("Failed to load data!")
                    exit()

            else:  # cache was not enabled
                if not cls._tmp_disabled:
                    cls._enable_default_cache()
                return func(api_path, **func_kwargs)

        return _cached_api_request

    @classmethod
    def _get_cache_file_path(cls, api_path, name):
        # extend the cache dir path using the api path and a file name
        # leading '/static/' is dropped form api path
        cache_dir_path = os.path.join(cls._CACHE_DIR, api_path[8:])
        if not os.path.exists(cache_dir_path):
            # create subfolders if they don't yet exist
            os.makedirs(cache_dir_path)

        file_name = name + '.ff1pkl'
        cache_file_path = os.path.join(cache_dir_path, file_name)
        return cache_file_path

    @classmethod
    def _data_ok_for_use(cls, cached):
        # check if cached data is ok or needs to be downloaded again
        if cls._FORCE_RENEW:
            return False
        elif cls._IGNORE_VERSION:
            return True
        elif cached['version'] == cls._API_CORE_VERSION:
            return True
        return False

    @classmethod
    def _write_cache(cls, data, cache_file_path, **kwargs):
        new_cached = dict(
            **{'version': cls._API_CORE_VERSION, 'data': data},
            **kwargs
        )
        with open(cache_file_path, 'wb') as cache_file_obj:
            pickle.dump(new_cached, cache_file_obj)

    @classmethod
    def _get_default_cache_path(cls):
        if sys.platform == "linux":
            # If .cache exists we will use it. Otherwise, ~/
            tmp = os.path.expanduser("~/.cache")
            if os.path.exists(tmp):
                return r"~/.cache/fastf1"
            else:
                return r"~/.fastf1"
        elif sys.platform == "darwin":
            return r"~/Library/Caches/fastf1"
        elif sys.platform == "win32":
            return r"%LOCALAPPDATA%\Temp\fastf1"
        else:
            return None

    @classmethod
    def _enable_default_cache(cls):
        if not cls._CACHE_DIR and not cls._default_cache_enabled:
            cache_dir = None
            if "FASTF1_CACHE" in os.environ:
                cache_dir = os.environ.get("FASTF1_CACHE")
            else:
                cache_dir = cls._get_default_cache_path()

            if cache_dir is not None:
                # Ensure the default cache folder exists
                cache_dir = os.path.expandvars(cache_dir)
                cache_dir = os.path.expanduser(cache_dir)
                if not os.path.exists(cache_dir):
                    try:
                        os.mkdir(cache_dir, mode=0o0700)
                    except Exception as err:
                        _logger.error("Failed to create cache directory {0}. "
                                      "Error {1}".format(cache_dir, err))
                        raise

                # Enable cache with default
                cls.enable_cache(cache_dir)
                _logger.warning(
                    f"DEFAULT CACHE ENABLED! "
                    f"({cls._convert_size(cls._get_size(cache_dir))}) "
                    f"{cache_dir}"
                )
            else:
                # warn only once and only if cache is not enabled
                _logger.warning(
                    "\n\nNO CACHE! Api caching has not been enabled! \n\t"
                    "It is highly recommended to enable this feature for much "
                    "faster data loading!\n\t"
                    "Use `fastf1.Cache.enable_cache('path/to/cache/')`\n")

                cls._default_cache_enabled = True

    @classmethod
    def disabled(cls):
        """Returns a context manager object that creates a context within
        which the cache is temporarily disabled.

        Example::

            with Cache.disabled():
                # no caching takes place here
                ...

        .. note::
            The context manager is not multithreading-safe
        """
        return _NoCacheContext()

    @classmethod
    def set_disabled(cls):
        """Disable the cache while keeping the configuration intact.

        This disables stage 1 and stage 2 caching!

        You can enable the cache at any time using :func:`set_enabled`

        .. note:: You may prefer to use :func:`disabled` to get a context
            manager object and disable the cache only within a specific
            context.

        .. note::
            This function is not multithreading-safe
        """
        cls._tmp_disabled = True

    @classmethod
    def set_enabled(cls):
        """Enable the cache after it has been disabled with
        :func:`set_disabled`.

        .. warning::
            To enable the cache it needs to be configured properly. You need
            to call :func`enable_cache` once to enable the cache initially.
            :func:`set_enabled` and :func:`set_disabled` only serve to
            (temporarily) disable the cache for specific parts of code that
            should be run without caching.

        .. note::
            This function is not multithreading-safe
        """
        cls._tmp_disabled = False

    @classmethod
    def offline_mode(cls, enabled: bool):
        """Enable or disable offline mode.

        In this mode, no actual requests will be sent and only cached data is
        returned. This can be useful for freezing the state of the cache or
        working with an unstable internet connection.

        Args:
            enabled: sets the state of offline mode to 'enabled' (``True``)
                or 'disabled' (``False``)
        """
        if cls._requests_session_cached is None:
            cls._enable_default_cache()
        cls._requests_session_cached.settings.only_if_cached = enabled

    @classmethod
    def ci_mode(cls, enabled: bool):
        """Enable or disable CI mode.

        In this mode, cached requests will be reused even if they are expired.
        Only uncached data will actually be requested and is then cached. This
        means, as long as CI mode is enabled, every request is only ever made
        once and reused indefinitely.

        This serves two purposes. First, reduce the number of requests that is
        sent on when a large number of tests is run in parallel, potentially
        in multiple environments simultaneously. Second, make test runs more
        predictable because data usually does not change between runs.

        Additionally, the pickle cache (stage 2) is disabled completely, so
        no parsed data is cached. This means that the API parser code is
        always executed and not skipped due to caching.
        """
        cls._ci_mode = enabled

    @classmethod
    def get_cache_info(cls) -> Tuple[Optional[str], Optional[int]]:
        """Returns information about the cache directory and its size.

        If the cache is not configured, None will be returned for both the
        cache path and the cache size.

        Returns:
            A tuple of ``(path, size)`` if the cache is configured, else
            ``(None, None)``. The cache size is given in bytes.
        """
        path = cls._CACHE_DIR
        if path is not None:
            size = cls._get_size(path)
        else:
            size = None

        return path, size

    @classmethod
    def _convert_size(cls, size_bytes):  # https://stackoverflow.com/questions/5194057/better-way-to-convert-file-sizes-in-python # noqa: E501
        if size_bytes == 0:
            return "0B"
        size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
        i = int(math.floor(math.log(size_bytes, 1024)))
        p = math.pow(1024, i)
        s = round(size_bytes / p, 2)
        return "%s %s" % (s, size_name[i])

    @classmethod
    def _get_size(cls, start_path='.'):  # https://stackoverflow.com/questions/1392413/calculating-a-directorys-size-using-python # noqa: E501
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(start_path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                # skip if it is symbolic link
                if not os.path.islink(fp):
                    total_size += os.path.getsize(fp)

        return total_size


class _NoCacheContext:
    def __enter__(self):
        Cache.set_disabled()

    def __exit__(self, exc_type, exc_val, exc_tb):
        Cache.set_enabled()


# TODO: document
class RateLimitExceededError(Exception):
    """Raised if a hard rate limit is exceeded."""
    pass



"""Driver Laptimes Scatterplot
==============================

Plot a driver's lap times in a race, with color coding for the compounds.
"""

import seaborn as sns
from matplotlib import pyplot as plt

import fastf1
import fastf1.plotting


# Enable Matplotlib patches for plotting timedelta values and load
# FastF1's dark color scheme
fastf1.plotting.setup_mpl(mpl_timedelta_support=True, misc_mpl_mods=False,
                          color_scheme='fastf1')


###############################################################################
# Load the race session.

race = fastf1.get_session(2023, "Azerbaijan", 'R')
race.load()

###############################################################################
# Get all the laps for a single driver.
# Filter out slow laps as they distort the graph axis.

driver_laps = race.laps.pick_driver("ALO").pick_quicklaps().reset_index()

###############################################################################
# Make the scattterplot using lap number as x-axis and lap time as y-axis.
# Marker colors correspond to the compounds used.
# Note: as LapTime is represented by timedelta, calling setup_mpl earlier
# is required.

fig, ax = plt.subplots(figsize=(8, 8))

sns.scatterplot(data=driver_laps,
                x="LapNumber",
                y="LapTime",
                ax=ax,
                hue="Compound",
                palette=fastf1.plotting.get_compound_mapping(session=race),
                s=80,
                linewidth=0,
                legend='auto')
# sphinx_gallery_defer_figures

###############################################################################
# Make the plot more aesthetic.
ax.set_xlabel("Lap Number")
ax.set_ylabel("Lap Time")

# The y-axis increases from bottom to top by default
# Since we are plotting time, it makes sense to invert the axis
ax.invert_yaxis()
plt.suptitle("Alonso Laptimes in the 2023 Azerbaijan Grand Prix")

# Turn on major grid lines
plt.grid(color='w', which='major', axis='both')
sns.despine(left=True, bottom=True)

plt.tight_layout()
plt.show()



"""Team Pace Comparison
=============================================
Rank team's race pace from the fastest to the slowest.
"""

import seaborn as sns
from matplotlib import pyplot as plt

import fastf1
import fastf1.plotting


# Load FastF1's dark color scheme
fastf1.plotting.setup_mpl(mpl_timedelta_support=False, misc_mpl_mods=False,
                          color_scheme='fastf1')


###############################################################################
# Load the race session.
# Pick all quick laps (within 107% of fastest lap).
# For races with mixed conditions, pick_wo_box() is better.
race = fastf1.get_session(2024, 1, 'R')
race.load()
laps = race.laps.pick_quicklaps()

###############################################################################
# Convert the lap time column from timedelta to integer.
# This is a seaborn-specific modification.
# If plotting with matplotlib, set mpl_timedelta_support to true
# with plotting.setup_mpl.
transformed_laps = laps.copy()
transformed_laps.loc[:, "LapTime (s)"] = laps["LapTime"].dt.total_seconds()

# order the team from the fastest (lowest median lap time) tp slower
team_order = (
    transformed_laps[["Team", "LapTime (s)"]]
    .groupby("Team")
    .median()["LapTime (s)"]
    .sort_values()
    .index
)
print(team_order)

# make a color palette associating team names to hex codes
team_palette = {team: fastf1.plotting.get_team_color(team, session=race)
                for team in team_order}

###############################################################################
fig, ax = plt.subplots(figsize=(15, 10))
sns.boxplot(
    data=transformed_laps,
    x="Team",
    y="LapTime (s)",
    hue="Team",
    order=team_order,
    palette=team_palette,
    whiskerprops=dict(color="white"),
    boxprops=dict(edgecolor="white"),
    medianprops=dict(color="grey"),
    capprops=dict(color="white"),
)

plt.title("2023 British Grand Prix")
plt.grid(visible=False)

# x-label is redundant
ax.set(xlabel=None)
plt.tight_layout()
plt.show()



"""Overlaying speed traces of two laps
======================================

Compare two fastest laps by overlaying their speed traces.
"""


import matplotlib.pyplot as plt

import fastf1.plotting


# Enable Matplotlib patches for plotting timedelta values and load
# FastF1's dark color scheme
fastf1.plotting.setup_mpl(mpl_timedelta_support=True, misc_mpl_mods=False,
                          color_scheme='fastf1')

# load a session and its telemetry data
session = fastf1.get_session(2021, 'Spanish Grand Prix', 'Q')
session.load()

##############################################################################
# First, we select the two laps that we want to compare

ver_lap = session.laps.pick_driver('VER').pick_fastest()
ham_lap = session.laps.pick_driver('HAM').pick_fastest()

##############################################################################
# Next we get the telemetry data for each lap. We also add a 'Distance' column
# to the telemetry dataframe as this makes it easier to compare the laps.

ver_tel = ver_lap.get_car_data().add_distance()
ham_tel = ham_lap.get_car_data().add_distance()

##############################################################################
# Finally, we create a plot and plot both speed traces.
# We color the individual lines with the driver's team colors.

rbr_color = fastf1.plotting.get_team_color(ver_lap['Team'], session=session)
mer_color = fastf1.plotting.get_team_color(ham_lap['Team'], session=session)

fig, ax = plt.subplots()
ax.plot(ver_tel['Distance'], ver_tel['Speed'], color=rbr_color, label='VER')
ax.plot(ham_tel['Distance'], ham_tel['Speed'], color=mer_color, label='HAM')

ax.set_xlabel('Distance in m')
ax.set_ylabel('Speed in km/h')

ax.legend()
plt.suptitle(f"Fastest Lap Comparison \n "
             f"{session.event['EventName']} {session.event.year} Qualifying")

plt.show()
